#!/usr/bin/env python

# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import os
import os.path
import re
import datetime
import distutils.spawn
import subprocess
import HTMLParser

from optparse import OptionParser

import reschema
import reschema.html
from reschema import ServiceDef, ServiceDefManager
from reschema.tohtml import ServiceDefToHtml, Options


class ReschemaDocException(Exception):
    def __init__(self, msg):
        super(ReschemaDocException, self).__init__()
        self.msg = msg


class ReschemaDoc(object):
    def __init__(self):
        self.servicedefmgr = ServiceDefManager()

    def parse_args(self, args):
        self.parser = parser = OptionParser(
            description="Generate docuemnation for service defintions.")
        parser.add_option('-f', '--file', dest='filename',
                          help='JSON doc source file', action="store")

        parser.add_option('-r', '--related', dest='related',
                          help='JSON doc source file', action="append")

        parser.add_option('-o', '--outdir', dest='outdir', default=None,
                          help='Output directory', action="store")

        parser.add_option('--html', dest='html', default=False,
                          help="Generate HTML", action="store_true")

        parser.add_option('--pdf', dest='pdf', default=False,
                          help="Generate PDF", action="store_true")

        parser.add_option('-w', '--wkhtmltopdf', dest='wkhtmltopdf', default=None,
                          help='Path to wkhtmltopdf tool', action="store")

        parser.add_option('--copyright', dest='copyright', default=None,
                          help='Company and years for copyright, '
                          'eg "Riverbed Technology Inc. 2014"', action="store")

        parser.add_option('--nojson', action="store_true", default=False,
                          help='Do not include JSON tabs')

        parser.add_option('--noxml', action="store_true", default=False,
                          help='Do not include XML tabs')

        parser.add_option('--device', dest='device', default='{device}',
                          help='Device hosting this service')

        parser.add_option('--apiroot', dest='apiroot', default=None,
                          help='Root path for all resources')

        parser.add_option('--urlname', default=None,
                          help='Name of this service in the URL (if different than name)')

        (options, args) = parser.parse_args(args)

        if not options.filename:
            raise ReschemaDocException("-f FILENAME is required")

        if not options.outdir:
            raise ReschemaDocException("-o OUTDIR is required")

        if not options.pdf and not options.html:
            raise ReschemaDocException(
                "Output format of --pdf and/or --html must be specified")

        self.options = options

    def run(self):
        options = self.options

        servicedef = ServiceDef()
        servicedef.load(options.filename)
        self.servicedefmgr.add(servicedef)

        for related in (options.related or []):
            relateddef = ServiceDef()
            relateddef.load(related)
            self.servicedefmgr.add(relateddef)

        title = "%s v%s %s" % (servicedef.title, servicedef.version,
                               servicedef.status)

        name = (self.options.urlname or servicedef.name)
        relname = os.path.join(name, servicedef.version, 'service')

        outdir = options.outdir

        fullname = os.path.abspath(os.path.join(outdir, relname))
        fulldir = os.path.dirname(fullname)

        if not os.path.exists(fulldir):
            os.makedirs(fulldir)

        apiroot = options.apiroot or ("/api/%s/%s" %
                                      (servicedef.name, servicedef.version))

        # HTML version
        if options.html:
            html = fullname + '.html'
            if os.path.exists(html):
                os.remove(html)

            h = HTMLParser.HTMLParser()

            htmldoc = reschema.html.Document(title, printable=False)
            htmldoc.header.a(href="http://www.riverbed.com", cls="headerimg")
            hl = htmldoc.header.div(cls="headerleft")
            breadcrumbs = hl.div(cls="breadcrumbs")
            breadcrumbs.a(href="../../index.html").text = "apis"
            breadcrumbs.span().text = h.unescape(" &raquo; ")
            breadcrumbs.a(href=("../index.html")).text = name
            breadcrumbs.span().text = h.unescape(" &raquo; %s" % servicedef.version)
            hl.div(cls="headertitle").text = title
            htmldoc.header.span(cls="headerright").text = (
                "Created %s" %
                datetime.datetime.now().strftime("%b %d, %Y at %I:%M %p"))

            r2h = ServiceDefToHtml(servicedef, htmldoc.content, htmldoc.menu,
                                   device=options.device,
                                   apiroot=apiroot,
                                   options=Options(printable=False,
                                                   json=(not options.nojson),
                                                   xml=(not options.noxml),
                                                   apiroot=apiroot,
                                                   docroot=outdir))
            r2h.process()
            htmldoc.write(html)
            print "Wrote %s" % html

        # PDF
        if options.pdf:
            # First make a printable HTML
            phtml = fullname + '-printable.html'
            if os.path.exists(phtml):
                os.remove(phtml)

            htmldoc = reschema.html.Document(title, printable=True)
            r2h = ServiceDefToHtml(servicedef, htmldoc.content, htmldoc.menu,
                                   device=options.device,
                                   apiroot=apiroot,
                                   options=Options(printable=True,
                                                   json=(not options.nojson),
                                                   xml=(not options.noxml),
                                                   apiroot=apiroot,
                                                   docroot=outdir))
            r2h.process()
            htmldoc.write(phtml)
            print "Wrote %s" % phtml

            ### PDF
            if options.wkhtmltopdf is not None:
                wkhtmltopdf = options.wkhtmltopdf
            else:
                wkhtmltopdf = distutils.spawn.find_executable("wkhtmltopdf")
                if not wkhtmltopdf:
                    if not "WKHTMLTOPDF" in os.environ:
                        raise ReschemaDocException(
                            "Cannot find 'wkhtmltopdf' in path, "
                            "use -w <path> or set WKHTMLTOPDF env variable")
                    wkhtmltopdf = os.environ['WKHTMLTOPDF']

            args = [wkhtmltopdf, "--version"]
            lines = subprocess.check_output(args).split('\n')
            version = None
            for line in lines:
                g = re.search("wkhtmltopdf ([0-9]+)\.([0-9]+)", line)
                if g:
                    version = int(g.group(2))
                    break

            if version is None:
                print ("WARNING: Could not determine wkhtmltopdf version, "
                       "assuming latest")
            else:
                print "wkhtmltopdf version %d" % version

            if version is None or version >= 10:
                tocarg = "toc"
                coverarg = "cover"
            else:
                tocarg = "--toc"
                coverarg = "--cover"

            # create a cover apge
            cover = fullname + '-cover.html'
            cover_base = reschema.html.HTMLElement('html')
            cover_body = cover_base.body()
            cover_body.h1().text = title
            if options.copyright:
                cover_body.p().text = u"Copyright \xa9 " + unicode(options.copyright)
            created = datetime.datetime.now().strftime("%b %d, %Y at %I:%m %p")
            cover_body.p().text = "Created %s" % created

            f = open(cover, "w")
            f.write(reschema.html.ET.tostring(cover_base, method="html"))
            f.close()

            pdf = fullname + '.pdf'
            if os.path.exists(pdf):
                os.remove(pdf)

            args = [wkhtmltopdf,
                    '--title', title,
                    coverarg, cover,
                    tocarg, '--toc-header-text', 'Contents',
                    '--footer-center', '[page]',
                    phtml, pdf]
            #print ' '.join(args)
            subprocess.check_call(args)
            # Fix for 'illegal byte sequence'
            os.environ['LC_TYPE'] = 'C'
            os.environ['LANG'] = 'C'
            args = ['sed', '-i.bak', '-e', 's/#00//g', pdf]
            subprocess.check_call(args)
            print "Wrote %s" % pdf
            os.remove(pdf + ".bak")



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

#pylint: disable-all

# Inspired By Dag Sverre Seljebotn
# https://gist.github.com/dagss/5008118

# OrderedDict method inspired by
# http://stackoverflow.com/questions/5121931/
"""
A PyYAML loader that annotates position in source code and uses an OrderedDict.

The loader is based on `SafeConstructor`, i.e., the behaviour of
`yaml.safe_load`, but in addition:

 - Every dict/list/unicode is replaced with dict_node/list_node/unicode_node,
   which subclasses dict/list/unicode to add the attributes `start_mark`
   and `end_mark`. (See the yaml.error module for the `Mark` class.)

 - Every string is always returned as unicode, no ASCII-ficiation is
   attempted.

 - Note that int/bool/... are returned unchanged for now
"""
from collections import OrderedDict

import yaml
from yaml.composer import Composer
from yaml.reader import Reader
from yaml.scanner import Scanner
from yaml.resolver import Resolver
from yaml.parser import Parser
from yaml.constructor import SafeConstructor, ConstructorError

from reschema.loader_nodes import (dict_node, list_node, unicode_node,
                                   add_marks_to_node)


class OrderedNodeConstructor(SafeConstructor):
    # To support lazy loading, the original constructors first yield
    # an empty object, then fill them in when iterated. Due to
    # laziness we omit this behaviour (and will only do "deep
    # construction") by first exhausting iterators, then yielding
    # copies.
    def _construct_ordereddict(self, node):
        # Inspired by http://stackoverflow.com/questions/5121931/
        mapping = OrderedDict()
        yield mapping
        if not isinstance(node, yaml.MappingNode):
            raise ConstructorError("while constructing an ordered map",
                                   node.start_mark,
                                   ("expected a mapping node, but found %s" %
                                    node.id),
                                   node.start_mark)
        for key_node, value_node in node.value:
            key = self.construct_object(key_node)
            try:
                hash(key)
            except TypeError, exc:
                raise ConstructorError('while constructing a mapping',
                                       node.start_mark,
                                       'found unacceptable key (%s)' % exc,
                                       key_node.start_mark)
            value = self.construct_object(value_node)
            mapping[key] = value

    def construct_yaml_omap(self, node):
        obj, = self._construct_ordereddict(node)
        return obj

OrderedNodeConstructor.add_constructor(
    'tag:yaml.org,2002:map', OrderedNodeConstructor.construct_yaml_omap)


class OrderedLoader(Reader, Scanner, Parser,
                   Composer, OrderedNodeConstructor, Resolver):
    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)

def ordered_load(stream):
    return OrderedLoader(stream).get_single_data()


class MarkedNodeConstructor(OrderedNodeConstructor):

    def construct_yaml_omap(self, node):
        obj, = self._construct_ordereddict(node)
        add_marks_to_node(obj, node.start_mark, node.end_mark)
        return obj

    def construct_yaml_map(self, node):
        obj, = SafeConstructor.construct_yaml_map(self, node)
        return dict_node(obj, node.start_mark, node.end_mark)

    def construct_yaml_seq(self, node):
        obj, = SafeConstructor.construct_yaml_seq(self, node)
        return list_node(obj, node.start_mark, node.end_mark)

    def construct_yaml_str(self, node):
        obj = SafeConstructor.construct_scalar(self, node)
        return unicode_node(obj, node.start_mark, node.end_mark)

MarkedNodeConstructor.add_constructor(
    'tag:yaml.org,2002:map', MarkedNodeConstructor.construct_yaml_omap)

MarkedNodeConstructor.add_constructor(
    'tag:yaml.org,2002:seq', MarkedNodeConstructor.construct_yaml_seq)

MarkedNodeConstructor.add_constructor(
    'tag:yaml.org,2002:str', MarkedNodeConstructor.construct_yaml_str)


class MarkedLoader(Reader, Scanner, Parser,
                   Composer, MarkedNodeConstructor, Resolver):
    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        SafeConstructor.__init__(self)
        Resolver.__init__(self)


def marked_load(stream):
    return MarkedLoader(stream).get_single_data()


def test_ordered_yaml():

    # note: test very sensitive to whitespace in string below
    d = ordered_load('''\
    a:
      [b, c, {d: e}]
    f:
      g: h''')

    assert d == {'a': ['b', 'c', {'d': 'e'}], 'f': {'g': 'h'}}

    assert isinstance(d['a'][2]['d'], str)
    assert isinstance(d, dict)
    assert isinstance(d['f'], dict)
    assert isinstance(d['a'], list)


def test_marked_yaml():
    def loc(obj):
        return (obj.start_mark.line, obj.start_mark.column,
                obj.end_mark.line, obj.end_mark.column)

    # note: test very sensitive to whitespace in string below
    d = marked_load('''\
    a:
      [b, c, {d: e}]
    f:
      g: h''')

    assert d == {'a': ['b', 'c', {'d': 'e'}], 'f': {'g': 'h'}}
    assert loc(d['a'][2]['d']) == (1, 17, 1, 18)
    assert loc(d) == (0, 4, 3, 10)
    assert loc(d['a']) == (1, 6, 1, 20)

    assert isinstance(d['a'][2]['d'], str)
    assert isinstance(d, dict)
    assert isinstance(d['f'], dict)
    assert isinstance(d['a'], list)


if __name__ == '__main__':
    test_ordered_yaml()
    test_marked_yaml()



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import re
import sys
import copy
import json
import yaml
import markdown
from collections import OrderedDict
import xml.dom.minidom, xml.etree.ElementTree as ET
import urlparse
import jsonpointer

import logging
logger = logging.getLogger(__name__)

import reschema
from reschema.html import HTMLElement, HTMLTable, Document as HTMLDoc, TabBar
from reschema.jsonschema import Schema
from reschema.util import a_or_an, str_to_id as html_str_to_id
from reschema.exceptions import NoManager


class Options(object):
    def __init__(self, printable=False, json=True, xml=True,
                 apiroot="/{root}", docroot='.'):
        self.printable = printable
        self.json = json
        self.xml = xml
        self.apiroot = apiroot
        self.docroot = docroot


class RefSchemaProxy(object):

    def __init__(self, schema, options=None, attrname='refschema'):
        self.schema = schema
        self.options = options

        try:
            self.passthrough = True
            self.refschema = getattr(schema, attrname)
            refid = self.refschema.fullid()
            self.name = self.refschema.name
            self.typestr = self.refschema.typestr
        except NoManager:
            self.passthrough = False
            refid = getattr(schema, '_' + attrname + '_id')
            self.name = refid.split('/')[-1]
            self.typestr = '<ref>'
            self.description = refid

        # refid is something like:
        #   http://support.riverbed.com/apis/test/1.0#/types/type_number_limits
        #
        # Drop the netloc and api root and replace with a relative path ref
        # based on the current schema id
        parsed_id = urlparse.urlparse(refid)

        # Fall back to just using an href of the schema id for the
        # following cases:
        #  - not a riverbed service
        #  - no match to '/apis'
        #  - looking for printable format
        m = re.match("/apis/(.*)$", parsed_id.path)
        if (parsed_id.netloc != 'support.riverbed.com' or
                not m or self.options.printable):
            self.href = refid
        else:
            # Build a relative link based on the difference between this
            # schema id and the refschema id.
            #
            # Figure out how many levels up to recurse -- up to the first diff
            parsed_parent_id = urlparse.urlparse(schema.servicedef.id)
            m_parent = re.match("/apis/(.*)$", parsed_parent_id.path)
            relpath_count = len(m_parent.group(1).split('/'))
            relpath = '/'.join(['..' for i in range(relpath_count)])

            frag_id = html_str_to_id(parsed_id.fragment)
            self.href = ('%s/%s/service.html#%s' %
                         (relpath, m.group(1), frag_id))

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]

        if self.__dict__['passthrough']:
            return getattr(self.refschema, name)
        else:
            return None


class ServiceDefToHtml(object):
    servicedef = None

    def __init__(self, servicedef, container, menu=None, options=None,
                 device="{device}", apiroot="/{root}"):
        ServiceDefToHtml.servicedef = servicedef
        self.servicedef = servicedef
        self.container = container
        self.menu = menu
        self.options = (options or Options())
        self.servicepath = "http://%s%s" % (device, apiroot)

    def process(self):

        resources_div = self.container.div(id='resources')
        self.menu.add_item("Resources", href=resources_div)
        resource_menu = self.menu.add_submenu()
        for resource in self.servicedef.resources.values():
            ResourceToHtml(resource, self.container, resource_menu,
                           self.servicepath, self.options).process()

        types_div = self.container.div(id='types')
        self.menu.add_item("Types", href=types_div)
        type_menu = self.menu.add_submenu()
        for type_ in self.servicedef.types.values():
            rh = ResourceToHtml(type_, self.container, type_menu,
                                self.servicepath, self.options)
            rh.process(is_type=True)


class ResourceToHtml(object):
    def __init__(self, schema, container, menu=None, basepath="",
                 options=None):
        self.schema = schema
        self.menu = menu
        self.container = container
        self.basepath = basepath
        self.options = (options or Options())

    def read(self, filename, pointer=None, additional=None):
        if filename == "-":
            f = sys.stdin
        else:
            f = open(filename, "r")

        if filename.endswith('.json'):
            input = json.load(f, object_pairs_hook=OrderedDict)
        elif filename.endswith(('.yml', '.yaml')):
            input = yaml.load(f)
        else:
            raise ValueError(
              "Unrecognized file extension, use '*.json' or '*.yml': %s" %
              filename)

        if additional:
            for a in additional:
                a_input = jsonpointer.resolve_pointer(input, pointer)
                Schema.parse(a_input, name=a)

        name = 'root'
        if pointer:
            input = jsonpointer.resolve_pointer(input, pointer)
            name = pointer

        self.schema_raw = input
        self.schema = Schema.parse(input, name=name)

        if f is not sys.stdin:
            f.close()

    def process(self, is_type=False):
        logger.debug("Processing resource: %s" % self.schema.fullname())
        menu = self.menu
        schema = self.schema
        baseid = html_str_to_id(schema.fullid(True))
        div = self.container.div(id=baseid)
        menu.add_item(schema.name, href=div)

        div.h2().text = ('Type: ' if is_type else 'Resource: ') + schema.name
        if schema.description != '':
            div.p().text = schema.description

        if (not is_type):
            uri = schema.links['self'].path.template
            if uri[0] == '$':
                uri = self.basepath + uri[1:]
            div.pre().text = uri

        self.schema_table(schema, div, baseid)

        if (not is_type):
            div.h3(id="links").text = "Links"
            self.process_links(div, baseid)

            if self.schema.relations:
                div.h3(id="relations").text = "Relations"
                self.process_relations(div, baseid)


    def schema_table(self, schema, container, baseid):
        tabbar = TabBar(container, baseid + '-tabbar',
                        printable=self.options.printable)
        if self.options.json:
            tabbar.add_tab("JSON", baseid + '-json',
                           SchemaSummaryJson(schema, self.options))
        if self.options.xml:
            tabbar.add_tab("XML", baseid + '-xml',
                           SchemaSummaryXML(schema, self.options))
        # tabbar.add_tab("JSON Schema", "jsonschema",
        #               HTMLElement("pre",
        #                            text=json.dumps(self.schema_raw,
        #                           indent=2)))
        tabbar.finish()
        container.append(SchemaTable(schema, self.options))

    def process_links(self, container, containerid):
        schema = self.schema
        for name, link in schema.links.iteritems():
            if name == 'self':
                # need to police invalid $refs underneath params
                for k, v in link.path.var_schemas.iteritems():
                    if isinstance(v, reschema.jsonschema.Ref):
                        RefSchemaProxy(v, self.options)
                continue

            logger.debug("Processing link: %s - %s" %
                         (self.schema.fullname(), name))

            baseid = containerid + '-link-%s' % name
            div = container.div(id=baseid, cls="link-body")
            div.h4().text = link.schema.fullname() + ": " + link.name
            div.p().text = link.description

            uri = link.path.template
            if uri[0] == '$':
                uri = self.basepath + uri[1:]

            path = uri

            httpmethod = link.method
            if httpmethod == "GET":
                # Handle link.request as parameters
                if link.request is not None and \
                       type(link.request) != reschema.jsonschema.Null:
                    properties = link.request.properties
                    params = ["%s={%s}" % (param, properties[param].typestr)
                              for param in properties.keys()]
                    if len(params) != 0:
                        path += "?" + "&".join(params)

            div.pre().text = httpmethod + " " + path

            # if link.headers:
                # div.span(cls="h5").text = "HTTP Headers"
                # table = div.table(cls="paramtable")
                # div.append(ParamTable(link.headers))
                # table.row(["Name", "Type", "Description"], header=True)
                # for name in link.parameters:
                #    p = link.parameters[name]
                #    table.row([name, p.typestr, p.description])

            if httpmethod != "GET":
                if link.request is not None and \
                       type(link.request) != reschema.jsonschema.Null and \
                       httpmethod != "GET":

                    div.span(cls="h5").text = "Request Body"

                    # Need to look at the raw link._request to see if it's
                    # a ref, as link.request is auto-resolved thru refs
                    if type(link._request) is reschema.jsonschema.Ref:
                        p = div.p()

                        schema = RefSchemaProxy(link._request, self.options)
                        p.settext("Provide ",
                                  a_or_an(schema.name),
                                  " ",
                                  p.a(cls="jsonschema-type", href=schema.href,
                                      text=schema.name),
                                  " data object.")
                    elif type(link.request) is reschema.jsonschema.Data:
                        p = div.p()
                        if link.request.description is None:
                            description = "data"
                        else:
                            description = "the " + link.request.description
                            p.settext(("Provide a request body containing %s "
                                       "with content type ") % description,
                                      p.span(cls="content-attribute",
                                             text=link.request.content_type),
                                      ".")
                    else:
                        p = div.p().text = ("Provide a request body with the "
                                            "following structure:")
                        self.schema_table(link.request, div,
                                          baseid + '-request')

                elif httpmethod in ("PUT", "POST"):
                    div.span(cls="h5").text = "Request Body"
                    div.p().text = "Do not provide a request body."

            div.span(cls="h5").text = "Response Body"
            # Need to look at the raw link._response to see if it's
            # a ref, as link.response is auto-resolved thru refs
            schema = link._response
            if schema is not None and \
                   type(schema) is not reschema.jsonschema.Null:
                if type(schema) is reschema.jsonschema.Ref:
                    p = div.p()
                    schema = RefSchemaProxy(link._response, self.options)
                    p.settext("Returns ",
                              a_or_an(schema.name),
                              " ",
                              p.a(cls="jsonschema-type", href=schema.href, text=schema.name),
                              " data object.")
                elif type(schema) is reschema.jsonschema.Data:
                    p = div.p()
                    p.settext("On success, the server returns a response body "
                              "containing data with content type ",
                              p.span(cls="content-attribute",
                                     text=schema.content_type),
                              ".")
                else:
                    p = div.p().text = ("On success, the server returns a "
                                        "response body with the following "
                                        "structure:")
                    self.schema_table(link.response, div, baseid + '-response')

            else:
                div.p().text = ("On success, the server does not provide any "
                                "body in the responses.")

            # XXXCJ - this code may still be needed, but not yet functional
            if None and link.examples is not None:
                div.span(cls="h5").text = "Examples"
                for ex in link.examples:
                    text = ex['text']
                    html = markdown.markdown(text)
                    md_html = ET.XML("<div>" + html + "</div>")
                    div.append(md_html)

        return None

    def process_relations(self, container, containerid):
        schema = self.schema
        for name, relation in schema.relations.iteritems():

            logger.debug("Processing relation: %s - %s" %
                         (self.schema.fullname(), name))

            baseid = containerid + '-relation-%s' % name
            div = container.div(id=baseid, cls="relation-body")
            div.h4().text = relation.schema.fullname() + ": " + relation.name
            div.p().text = relation.description

            div.span(cls="h5").text = "Related resource"
            target_schema = RefSchemaProxy(relation, self.options,
                                           attrname='resource')
            p = div.p()
            p.settext(p.a(cls="jsonschema-type", href=target_schema.href,
                          text=target_schema.name))

            if relation.vars:
                div.span(cls="h5").text = "Variables"
                table = HTMLTable(cls="paramtable")
                table.row(["Related var", "Data value for replacement"],
                          header=True)
                for var, relp in relation.vars.iteritems():
                    table.row([var, relp])
                div.append(table)


class SchemaSummaryJson(HTMLElement):
    def __init__(self, schema, options):
        HTMLElement.__init__(self, "pre", cls="servicedef")
        self.options = options
        self.process(self, schema, 0)

        if isinstance(schema, reschema.jsonschema.Ref):
            example = schema.refschema.example
        else:
            example = schema.example

        if example is not None:
            self.span().text = ("\n\nExample:\n%s\n" %
                                json.dumps(example, indent=2))

    def process(self, parent, schema, indent=0):
        if isinstance(schema, reschema.jsonschema.Object):
            self.process_object(parent, schema, indent)

        elif isinstance(schema, reschema.jsonschema.Array):
            self.process_array(parent, schema, indent)

        elif isinstance(schema, reschema.jsonschema.Ref):
            schema = RefSchemaProxy(schema, self.options)
            parent.a(cls="servicedef-type", href=schema.href, text=schema.name)

        elif isinstance(schema, reschema.jsonschema.Merge):
            self.process(parent, schema.refschema, indent)

        else:
            parent.span(cls="servicedef-type").text = schema.typestr

    def process_object(self, parent, obj, indent):
        logger.debug("process_object: obj: %s" % obj.fullname())
        parent.span().text = "{\n"
        last = None
        for k in obj.properties:
            if last is not None:
                last.text = ",\n"

            txt = ('%*.*s"%s": ' % (indent + 2, indent + 2, "", k))
            parent.span(cls="servicedef-property").text = txt

            s = parent.span()
            self.process(s, obj.properties[k], indent + 2)
            last = parent.span()
            last.text = '\n'

        if obj.additional_properties:
            if last is not None:
                last.text = ",\n"
            if obj.additional_properties is True:
                txt = '%*.*s%s' % (indent + 2, indent + 2, "", 'prop')
                parent.span(cls="servicedef-type").text = txt
                parent.span().text = ": "
                parent.span(cls="servicedef-type").text = 'value'
            else:
                txt = '%*.*s%s' % (indent + 2, indent + 2, "",
                                   obj.additional_properties.name)
                parent.span(cls="servicedef-type").text = txt
                parent.span().text = ": "
                s = parent.span()
                self.process(s, obj.additional_properties, indent + 2)

            last = parent.span()
            last.text = '\n'

        parent.span().text = "%*.*s}" % (indent, indent, "")

    def process_array(self, parent, array, indent):
        item = array.children[0]
        if isinstance(item, reschema.jsonschema.Ref):
            s = parent.span()
            schema = RefSchemaProxy(item, self.options)
            s.settext("[ ", s.a(cls="servicedef-type",
                                href=schema.href, text=schema.name),
                       " ]")

        else:
            parent.span().text = "[\n%*.*s" % (indent + 2, indent + 2, "")
            self.process(parent.span(), item, indent + 2)
            parent.span().text = "\n%*.*s]" % (indent, indent, "")


class SchemaSummaryXML(HTMLElement):
    def __init__(self, schema, options):
        HTMLElement.__init__(self, "pre", cls="servicedef")
        self.options = options
        self.text = self.process(self, schema, 0)

        if type(schema) is reschema.jsonschema.Ref:
            example = schema.refschema.example
        elif schema.xmlExample is not None:
            # XXX/demmer hack to support types that don't have
            # GL4-compliant XML/JSON
            self.span().text = "\n\nExample:\n%s\n" % schema.xmlExample
            example = None
        else:
            example = schema.example

        if example is not None:
            example_xml = schema.toxml(example)

            x2 = xml.dom.minidom.parseString(ET.tostring(example_xml))
            xs = x2.toprettyxml(indent="  ").split("\n")

            if re.search("\?xml", xs[0]):
                # Remove the leading "xml" tag
                xs = xs[1:]

            example_str = "\n".join(xs)

            self.span().text = "\n\nExample:\n%s\n" % example_str


    def process(self, parent, schema, indent=0, name=None,
                json=None, key=None):
        if (not isinstance(schema, reschema.jsonschema.Ref) and
                schema.xmlSchema is not None):
            # XXX/demmer this is a big hack to support the fact that
            # one of the Shark REST handlers doesn't return
            # GL5-compliant output. The JSON is properly described by
            # the schema, the but XML is specified separately.
            def write_xmlschema(tag, spec, indent):
                parent.span().text = "%*s<" % (indent, "")
                parent.span(cls="xmlschema-element").text = tag
                first = True
                attr_indent = "\n%*s" % (indent + 2 + len(tag), "")

                if 'attributes' in spec:
                    for attr in spec['attributes']:
                        typestr = spec['attributes'][attr]
                        parent.span(cls="xmlschema-attribute").text = (
                          '%s%s' % (" " if first else attr_indent, attr))
                        parent.span().text = '='
                        parent.span(cls="xmlschema-type").text = typestr
                        first = False

                if 'children' in spec:
                    parent.span().text = ">\n"
                    for child_tag, child_spec in spec['children'].items():
                        write_xmlschema(child_tag, child_spec, indent + 2)
                    parent.span().text = "%*s</" % (indent, "")
                    parent.span(cls="xmlschema-element").text = tag
                    parent.span().text = ">\n"
                else:
                    parent.span().text = "/>\n"

            assert (len(schema.xmlSchema.keys()) == 1)
            tag = schema.xmlSchema.keys()[0]
            spec = schema.xmlSchema[tag]
            write_xmlschema(tag, spec, 0)
            return

        name = name or schema.name
        if name == "[]":
            name = "items"
        if type(schema) is reschema.jsonschema.Object:
            self.process_object(parent, schema, indent, name, key=key)

        elif isinstance(schema, reschema.jsonschema.Array):
            self.process_array(parent, schema, indent, name, key=key)

        elif isinstance(schema, reschema.jsonschema.Ref):
            schema = RefSchemaProxy(schema, self.options)
            if not name:
                name = schema.name
            if key:
                parent.span(cls="xmlschema-element").text = (
                  "%*s<%s key=string>" % (indent, "", name))
            else:
                parent.span(cls="xmlschema-element").text = (
                  "%*s<%s>" % (indent, "", name))

            parent.a(cls="xmlschema-type", href=schema.href, text=name)
            parent.span(cls="xmlschema-element").text = "</%s>\n" % (name)

        else:
            if key:
                parent.span(cls="xmlschema-element").text = (
                  "%*s<%s " % (indent, "", name))
                parent.span(cls="xmlschema-attribute").text = key
                parent.span().text = "="
                parent.span(cls="xmlschema-type").text = "string"
                parent.span().text = ">"
            else:
                parent.span(cls="xmlschema-element").text = (
                  "%*s<%s>" % (indent, "", name))
            parent.span(cls="xmlschema-type").text = "%s" % schema.typestr
            parent.span(cls="xmlschema-element").text = "</%s>\n" % (name)

    def process_object(self, parent, obj, indent, name, key=None):
        parent.span().text = "%*s<" % (indent, "")
        parent.span(cls="xmlschema-element").text = name
        subelems = obj.additional_properties
        first = True
        attr_indent = "\n%*s" % (indent + 2 + len(name), "")

        for k in obj.properties:
            prop = obj.properties[k]
            if not prop.is_simple():
                subelems = True
                continue
            parent.span(cls="xmlschema-attribute").text = (
              '%s%s' % (" " if first else attr_indent, k))
            parent.span().text = '='
            parent.span(cls="xmlschema-type").text = prop.typestr
            first = False

        if subelems:
            parent.span().text = ">\n"
        else:
            parent.span().text = "/>\n"

        for k in obj.properties:
            prop = obj.properties[k]
            if prop.is_simple():
                continue
            s = parent.span()
            self.process(s, prop, indent + 2, name=k)

        if obj.additional_properties is True:
            pass
        elif obj.additional_properties:
            s = parent.span()

            subobj = copy.copy(obj.additional_properties)
            keyname = subobj.xmlKeyName or 'key'
            try:
                subobj.properties = OrderedDict()
                json = {'id':keyname, 'type':'string',
                        'description':'property name'}
                subobj.properties[keyname] = subobj.schema.parse(json,
                                                            keyname,
                                                            subobj)
                for k in obj.additional_properties.props:
                    subobj.props[k] = obj.additional_properties.props[k]
            except:
                pass
            self.process(s, subobj, indent + 2, name=subobj.name, key=keyname)

        if subelems:
            parent.span().text = "%*s</" % (indent, "")
            parent.span(cls="xmlschema-element").text = name
            parent.span().text = ">\n"

    def process_array(self, parent, array, indent, name, key):
        parent.span().text = "%*s<" % (indent, "")
        parent.span(cls="xmlschema-element").text = name
        parent.span().text = ">\n"
        self.process(parent.span(), array.children[0], indent + 2,
                     name=array.children[0].name)
        parent.span().text = "%*s</" % (indent, "")
        parent.span(cls="xmlschema-element").text = name
        parent.span().text = ">\n"


class PropTable(HTMLTable):
    def __init__(self, data, options):
        self.options = options
        HTMLTable.__init__(self, cls="paramtable")

        logger.debug("PropTable")

        self.define_columns(["paramtable-propname",
                             "paramtable-proptype",
                             "paramtable-description",
                             "paramtable-notes"])
        self.row(["Property Name", "Type", "Description", "Notes"],
                 header=True)

        self.process(data)

    def process(self, data):
        pass

    def setname(self, elem, name):
        if not name:
            return
        limit = 40
        L = re.split("[.[]", name)
        line = 0
        for i in range(len(L)):
            if L[i][-1] == "]":
                text = "[" + L[i]
            else:
                text = L[i]
            if line + len(text) > limit:
                elem.br()
                line = 2
                elem.span(cls="servicedef-indent", text="")
            if i != len(L) - 1:
                if L[i + 1][-1] != "]":
                    text += "."

            elem.span(cls="servicedef-%s" % "basename" if i == 0
                                                       else "property",
                      text=text)
            line += len(text)

    def makerow(self, schema, name):
        if isinstance(schema, reschema.jsonschema.Merge):
            return self.makerow(schema.refschema, schema.refschema.fullname())

        tds = self.row(["", "", "", ""])

        logger.debug("Row: %s - %s " % (schema.fullname(), name))
        # To avoid very long nested structures from taking up too much
        # room in the table, force line breaks after a set number of
        # characters.

        self.setname(tds[0], name)

        if isinstance(schema, reschema.jsonschema.Ref):
            schema = RefSchemaProxy(schema, self.options)
            tds[1].a(cls="servicedef-type", href=schema.href, text="<" + schema.name + ">")
        else:
            tds[1].span(cls="servicedef-type", text="<" + schema.typestr + ">")

        tds[2].text = schema.description

        desctd = tds[3]
        parts = []

        if schema.readOnly:
            parts.append("Read-only")

        if (isinstance(schema, reschema.jsonschema.Object) and
             schema.required is not None and
             len(schema.required) > 0):
            parts.append("Required properties: [%s]" %
                         ', '.join(schema.required))

        if ((schema.parent is not None) and
             (isinstance(schema.parent, reschema.jsonschema.Object)) and
             ((schema.parent.required is None) or
              (schema.name not in schema.parent.required)) and
             (not (re.match("anyOf|allOf|oneOf|not", schema.name)))
             ):
            parts.append("Optional")

        if (isinstance(schema, reschema.jsonschema.Number) or
             isinstance(schema, reschema.jsonschema.Integer)):
            minimum = None
            if schema.minimum is not None:
                minimum = str(schema.minimum)
                if schema.exclusiveMinimum:
                    minimum = "(%s)" % str(minimum)

            maximum = None
            if schema.maximum is not None:
                maximum = str(schema.maximum)
                if schema.exclusiveMaximum:
                    maximum = "(%s)" % str(maximum)

            if (minimum is not None) and (maximum is not None):
                parts.append("Range: %s to %s" % (minimum, maximum))
            elif (minimum is not None):
                parts.append("Minimum %s" % (minimum))
            elif (maximum is not None):
                parts.append("Maximum %s" % (maximum))

        if isinstance(schema, reschema.jsonschema.Array):
            if schema.minItems is not None and \
                   schema.maxItems is not None:
                parts.append("%s-%s items" % (schema.minItems,
                                              schema.maxItems))
            elif schema.minItems is not None:
                parts.append("Minimum: %s items" % (schema.minItems))
            elif schema.maxItems is not None:
                parts.append("Maximum: %s items" % (schema.maxItems))

        if isinstance(schema, reschema.jsonschema.Timestamp):
            parts.append("Seconds since January 1, 1970")

        if hasattr(schema, 'default') and schema.default is not None:
            parts.append("Default is %s" % schema.default)

        if hasattr(schema, 'enum') and schema.enum is not None:
            parts.append("Values: " + ', '.join([str(x) for x in schema.enum]))

        if hasattr(schema, 'pattern') and schema.pattern is not None:
            parts.append("Pattern: '" + schema.pattern + "'")

        if hasattr(schema, 'notes') and schema.notes is not None:
            parts.append(schema.notes)

        desctd.settext('; '.join(parts))


class ParamTable(PropTable):
    def __init__(self, parameters):
        PropTable.__init__(self, parameters)

    def process(self, parameters):
        for p in parameters:
            self.makerow(parameters[p], p)


class SchemaTable(PropTable):
    def __init__(self, schema, options):
        PropTable.__init__(self, schema, options)

    def process(self, schema):
        self.makerow(schema, schema.fullname())

        if isinstance(schema, reschema.jsonschema.Merge):
            schema = schema.refschema

        for child in schema.children:
            if re.match("anyOf|allOf|oneOf|not", child.name):
                continue
            self.process(child)

        if isinstance(schema, reschema.jsonschema.Object):
            if schema.additional_properties is True:
                tds = self.row(["", "", "", ""])
                self.setname(tds[0], schema.fullname() + ".<prop>")
                tds[1].span(cls="servicedef-type").text = "<value>"
                tds[2].text = ("Additional properties may have "
                               "any property name and value")

        for child in schema.children:
            if not re.match("anyOf|allOf|oneOf|not", child.name):
                continue
            self.process(child)

if __name__ == '__main__':
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option('-f', '--file', dest='filename',
                      help='source file', action="store")
    parser.add_option('-o', '--output', dest='output',
                      help='HTML output file')
    parser.add_option('-p', '--pointer', dest='pointer',
                      help='JSON pointer to element within source to parse')
    parser.add_option('-a', '--additional', dest='additional', action="append",
                      help='JSON pointer to element within source to parse')

    (options, args) = parser.parse_args()

    htmldoc = HTMLDoc(options.filename)
    s2h = SchemaToHtml(basepath="/api", document=htmldoc)
    s2h.read(options.filename, options.pointer, options.additional)
    htmldoc.header.span(cls="headerleft").text = s2h.schema.name
    s2h.process()
    htmldoc.write(options.output)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import json
from collections import OrderedDict

from yaml.error import Mark
from json.scanner import py_make_scanner
import json.decoder

from reschema.loader_nodes import ordered_dict_node, list_node, unicode_node


def linecol(doc, pos):
    # assume zero-indexed
    lineno = doc.count('\n', 0, pos)
    if lineno == 0:
        colno = pos - 1
    else:
        colno = pos - doc.rindex('\n', 0, pos) - 1
    return lineno, colno


class Decoder(json.decoder.JSONDecoder):
    def __init__(self, name="", *args, **kwargs):
        super(Decoder, self).__init__(*args, **kwargs)

        def wrap_obj_parser(parser, node_type):
            def internal(o_and_start, *args, **kwargs):
                o, start = o_and_start
                r, end = parser(o_and_start, *args, **kwargs)

                start_line, start_col = linecol(o, start)
                end_line, end_col = linecol(o, end)
                start_mark = Mark(name, start, start_line, start_col, o, start)
                end_mark = Mark(name, end, end_line, end_col, o, start)

                return node_type(r, start_mark, end_mark), end
            return internal

        def wrap_parser(parser, node_type):
            def internal(o, start, *args, **kwargs):
                r, end = parser(o, start, *args, **kwargs)

                start_line, start_col = linecol(o, start)
                end_line, end_col = linecol(o, end)
                start_mark = Mark(name, start, start_line, start_col, o, start)
                end_mark = Mark(name, end, end_line, end_col, o, start)

                return node_type(r, start_mark, end_mark), end
            return internal

        self.parse_string = wrap_parser(self.parse_string, unicode_node)
        self.parse_array = wrap_obj_parser(self.parse_array, list_node)
        self.parse_object = wrap_obj_parser(self.parse_object,
                                            ordered_dict_node)

        # Not thread safe, but need to patch this for loading marks onto object
        # keys.
        json.decoder.scanstring = self.parse_string

        # Need to hook the python scanner because the C scanner doesn't have
        # a hookable method to parse_string.
        self.scan_once = py_make_scanner(self)


def marked_load(stream):
    return json.load(stream, cls=Decoder, name=stream.name)


def clean_load(stream):
    return json.load(stream, object_pairs_hook=OrderedDict)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from gitpy_versioning import get_version

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


readme = open('README.rst').read()

doc = [
    'sphinx',
]

test = [
    'pytest',
    'mock',
]

setup(
    name='reschema',
    version=get_version(),
    description=("reschema - Parse REST API service defintions into Python "
                 "objects, support documentation generation"),
    long_description=readme,
    author="Riverbed Technology",
    author_email="eng-github@riverbed.com",
    packages=[
        'reschema',
    ],
    package_dir={'reschema': 'reschema'},
    scripts=[
        'bin/reschema-doc',
        'bin/relint'
    ],
    include_package_data=True,
    install_requires=[
        'PyYAML',
        'Markdown',
        'uritemplate',
        'jsonpointer',
    ],
    extras_require={
        'test': test,
        'doc': doc,
        'dev': test + doc,
        'all': [],
    },
    tests_require=test,
    url="http://pythonhosted.org/steelscript",
    keywords='reschema',
    license='MIT',
    platforms='Linux, Mac OS, Windows',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Documentation',
    ],
)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import os

#
# Set to True to load all description fields associated with
# a JSON schema.  If false (default), the descriptions will be blank
#
LOAD_DESCRIPTIONS = ('RESCHEMA_LOAD_DESCRIPTIONS' in os.environ)

#
# Set to True to add file / line numbers while loading schema
# files.  Useful when debugging yaml/json file syntax errors
# Defaults to False
#
MARKED_LOAD = ('RESCHEMA_MARKED_LOAD' in os.environ)

#
# Set to True for verbose debugging
#
VERBOSE_DEBUG = ('RESCHEMA_VERBOSE_DEBUG' in os.environ)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import urlparse

from reschema.exceptions import ParseError, InvalidReference
from reschema.util import check_type

from reschema.loader_nodes import obj_key_node

import reschema.settings


class Parser(object):
    """ Input object parser.

    Parsers handle checking for required properties, checking the data
    type, managing a caller-supplied default value, and optionally
    creating a data member on the ``obj`` argument with the same name
    as the property, and setting it to the property value.

    A given Parser instance only ever works with the top level of
    the input dictionary.  No processing of values beyond checking
    their types and making use of any supplied default values is performed.

    Recursive traversal of the input should
    be handled by callers creating new parsers on dictionaries
    parsed out of the parent input.

    A Parser may be used as a context manager that ensures that all
    input dictionary keys have been consumed upon exit.

    :param dict input: input to parser
    :param name: name for logging / error
    :param obj: object on which to set attributes (i.e. the "context")

    If obj is None, it may be set later with set_context(), but
    obj may not be altered once set.
    """
    def __init__(self, input, name, obj=None, base_id=None):
        # TODO: Does any caller pass base_id?  The parameter is no
        #       longer actually used in this class.

        if not isinstance(input, dict):
            raise ParseError('%s: definition should be a dictionary, got: %s' %
                             (name, type(input)), input)

        self.input = input
        self.obj = None
        self.name = name
        self.parsed_props = set()

        if obj:
            self.set_context(name, obj)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if type is None:
            # parsing was all successful, check that all input keys
            # were parsed
            self.check_input()
            return True

    def set_context(self, name, obj):
        """Set the object to modify when parsing.

        :param name: name for logging / error
        :param obj: object on which to set attributes

        """
        if self.obj is not None:
            raise ParseError('Cannot change object context, '
                             'only set if it was None', input)

        self.obj = obj
        self.name = name

        self.mark_object(obj)

    def mark_object(self, obj, prop=None, as_key=False, required=False):
        """Associtate input property marks with a object.

        Used by tools that use marked input to copy their marks to
        representitive objects. Invoke for objects that are not constructed
        by the parser class.

        :param obj: The object that represents the property
        :param prop: The property that is being represented. If None, the
            full input is used.
        :param as_key: Assoicate with the key of the property instead of it's
            value.
        :param bool required: Causes ParseError to be raised if `prop`
            is not in the input
        :raises reschema.exceptions.ParseError: if the type of the
            data is incorrect or if the property is required but
            missing.

        """

        # If object aleady is marked, short circuit.
        if hasattr(obj, "start_mark"):
            return

        if prop is None:
            val = self.input
        else:
            val = obj_key_node(self.input, prop)
            if val is None:
                if required:
                    raise ParseError(
                        "Missing required property '%s'" % prop, self.input)
                return
            if not as_key:
                val = self.input[val]

        if hasattr(val, "start_mark"):
            obj.start_mark = val.start_mark
            obj.end_mark = val.end_mark

    def parse(self, prop, default_value=None, required=False,
              types=None, save=True, save_as=None):
        """Parse a key from the input dict.

        :param string prop: Property name to extract

        :param default_value: A value to use if the property is missing
            but not required.  Ignored if `required` is True.

        :param bool required: Causes ParseError to be raised if `prop`
            is not in the input

        :param list types: verifies that the property value is an
            instance of at least one of the types passed in this
            parameter.

        :param save: If true, save the property to this parsers
            object, if false, don't save (value is returned)

        :param str save_as: the property name to save as, defaults
            to ``prop``

        :raises reschema.exceptions.ParseError: if the type of the
            data is incorrect or if the property is required but
            missing.

        :return: The parsed value.
        """

        if (prop == 'description' and not reschema.settings.LOAD_DESCRIPTIONS):
            val = ''
            if prop in self.input:
                del(self.input[prop])
                self.parsed_props.add(prop)

        elif prop in self.input:
            val = self.input[prop]
            if types:
                check_type(prop, val, types, self.input)
            self.parsed_props.add(prop)

        elif required:
            raise ParseError(
                "Missing required property '%s'" % prop, self.input)
        else:
            val = default_value

        if save:
            if not self.obj:
                raise ParseError(
                    'Cannot save prop %s, no obj set' % prop, self.input)

            setattr(self.obj, save_as or prop, val)

        return val

    def check_input(self):
        """ Verify that all input properities were parsed. """
        if input is None:
            return

        unparsed = set(self.input.keys()).difference(self.parsed_props)

        if len(unparsed) > 0:
            raise ParseError(
                '%s: unrecognized properties in definition: %s' %
                (self.name, ','.join(unparsed)), list(unparsed)[0])

    @classmethod
    def expand_ref(cls, base_id, ref):
        """ Expand a reference using base_id as a relative base

        Returns a fully qualified reference based.

        :param reference: string reference to resolve

        The `reference` may be one of three supported forms:

           * `<server><path>#<fragment>` - fully qualified reference

           * `<path>#<fragment>` - reference is resolved against the
             same <server> as `base_id`.  <path> starts with '/'

           * `#<fragment>` - reference is resolved against the same
             <server> and <path> as `base_id`

        :raises InvalidReference: `reference` does not appear to
            be to the correct syntax

        """

        parsed_ref = urlparse.urlparse(ref)
        if parsed_ref.netloc:
            # Already a fully qualified address, let urlparse rejoin
            # to normalize it
            return parsed_ref.geturl()

        if ref[0] not in ['/', '#']:
            raise InvalidReference("relative references should "
                                   "start with '#' or '/'",
                                   ref)

        # urljoin will take care of the rest
        return urlparse.urljoin(base_id, ref)

    def preprocess_input(self, base_id):
        """Perform some preprocessing on our input."""
        self.preprocess(base_id, self.input)

    @classmethod
    def preprocess(cls, base_id, input):
        """Perform preprocessing on the supplied input."""
        cls.expand_refs(base_id, input)

    @classmethod
    def expand_refs(cls, base_id, input):
        """ Replace all relative refs in input with absolute refs"""

        if input:
            if isinstance(input, dict):
                if '$ref' in input and isinstance(input['$ref'], basestring):
                    # replace relative refs with fully expanded refs
                    oldref = input['$ref']
                    newref = cls.expand_ref(base_id, oldref)
                    input['$ref'] = newref

                elif len(input.keys()) > 0:
                    for k, v in input.iteritems():
                        cls.expand_refs(base_id, v)

            elif isinstance(input, list):
                for v in input:
                    cls.expand_refs(base_id, v)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from __future__ import print_function

import os
import unittest
import shutil
import logging
from reschema.util import str_to_id as html_str_to_id

logger = logging.getLogger(__name__)

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
SERVICE_DEF_TEST = os.path.join(TEST_PATH, 'service_test.yaml')
SERVICE_DEF_TEST_REF = os.path.join(TEST_PATH, 'service_test_ref.yaml')
SERVICE_DEF_BOOKSTORE = os.path.join(TEST_PATH, '../examples/bookstore.yaml')
SERVICE_DEF_INVALID_REF = os.path.join(TEST_PATH,
                                       'service_test_invalid_ref.yaml')

outdir = 'test_reschema_doc_output'
if os.path.exists(outdir):
    shutil.rmtree(outdir)

os.makedirs(outdir)

import reschema
from reschema import ServiceDef
from reschema.tohtml import ResourceToHtml, RefSchemaProxy
from reschema.doc import ReschemaDoc


def process_file(filename, *args):
    r = ReschemaDoc()
    rargs = ['-f', filename,
             '--outdir', outdir,
             '--html']

    for arg in args:
        rargs.extend(['-r', arg])

    r.parse_args(rargs)
    r.run()
    return


class TestReschema(unittest.TestCase):

    def test_service(self):
        process_file(SERVICE_DEF_TEST)

    def test_service_ref(self):
        process_file(SERVICE_DEF_TEST_REF, SERVICE_DEF_TEST)

    def test_service_bookstore(self):
        process_file(SERVICE_DEF_BOOKSTORE)


class TestReschemaInvalidRef(unittest.TestCase):

    def setUp(self):
        self.sd = ServiceDef()
        self.sd.load(SERVICE_DEF_INVALID_REF)

    def test_invalid_ref_in_property(self):
        """
        testing when one property's ref consists of undefined types,
        an invalid reference exception should be raised, such as below:
        properties:
           property: { $ref: '#/types/blah' }
        """
        with self.assertRaises(reschema.exceptions.InvalidReference):
            schema = self.sd.resources.values()[0].properties['name']
            RefSchemaProxy(schema, None)

    def test_invalid_ref_in_links(self):
        """
        testing when one property's ref consists of undefined resources,
        an invalid reference exception should be raised. such as below:
        properties:
        links:
          self: { path: '$/test_invalid_ref_in_lnks'}
          params:
             id:
                $ref: '#/types/does_not_exist'
        """
        with self.assertRaises(reschema.exceptions.InvalidReference):
            resource = self.sd.resources.values()[0]
            title = "%s v%s %s" % (self.sd.title, self.sd.version,
                                   self.sd.status)
            htmldoc = reschema.html.Document(title, printable=False)
            r2h = ResourceToHtml(resource, htmldoc.content,
                                 htmldoc.menu.add_submenu(),
                                 "http://{device}/{root}",
                                 None)
            baseid = html_str_to_id(r2h.schema.fullid(True))
            div = r2h.container.div(id=baseid)
            r2h.menu.add_item(r2h.schema.name, href=div)
            r2h.process_links(div, baseid)


if __name__ == '__main__':
    logging.basicConfig(filename='test.log', level=logging.DEBUG)
    unittest.main()



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import os
import logging
import unittest
import pytest
import urlparse

from yaml.error import MarkedYAMLError

import reschema

from reschema.exceptions import (ValidationError, NoManager,
                                 MissingParameter, ParseError,
                                 InvalidReference)

from reschema.jsonschema import (Object, Integer, String, Array, Schema)
from reschema import yaml_loader, ServiceDef, ServiceDefManager

logger = logging.getLogger(__name__)

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
SERVICE_DEF_TEST = os.path.join(TEST_PATH, 'service_test.yaml')
SERVICE_DEF_TEST_REF = os.path.join(TEST_PATH, 'service_test_ref.yaml')

PACKAGE_PATH = os.path.dirname(TEST_PATH)

EXAMPLES_DIR = os.path.join(PACKAGE_PATH, 'examples')
BOOKSTORE_YAML = os.path.join(EXAMPLES_DIR, 'bookstore.yaml')
BOOKSTORE_JSON = os.path.join(EXAMPLES_DIR, 'bookstore.json')


class TestReschema(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_load_schema(self):
        # Test marked load and dropping descriptions
        reschema.settings.MARKED_LOAD = True
        reschema.settings.LOAD_DESCRIPTIONS = False
        r = ServiceDef()
        r.load(BOOKSTORE_YAML)
        self.assertEqual(r.name, 'bookstore')
        self.assertEqual(r.check_references(), [])

        info = r.resources['info']
        self.assertTrue(hasattr(info.input, 'start_mark'))
        self.assertEqual(info.description, '')

        info_descr = info.by_pointer('/description')
        self.assertEqual(info_descr.default, 'Info Description')

        address = r.types['address']
        city = address.by_pointer('/city')
        self.assertEqual(city.description, '')

        # Test unmarked load and not loading descriptions (defaults)
        reschema.settings.MARKED_LOAD = False
        reschema.settings.LOAD_DESCRIPTIONS = True
        r = ServiceDef()
        r.load(BOOKSTORE_YAML)
        self.assertEqual(r.name, 'bookstore')
        self.assertEqual(r.check_references(), [])

        info = r.resources['info']
        self.assertFalse(hasattr(info.input, 'start_mark'))
        self.assertNotEqual(info.description, '')

        info_descr = info.by_pointer('/description')
        self.assertEqual(info_descr.default, 'Info Description')

        address = r.types['address']
        city = address.by_pointer('/city')
        self.assertNotEqual(city.description, '')

    def test_load_schema_json(self):
        reschema.settings.MARKED_LOAD = True
        r = ServiceDef()
        r.load(BOOKSTORE_JSON)
        self.assertEqual(r.name, 'bookstore')
        info = r.resources['info']
        self.assertTrue(hasattr(info.input, 'start_mark'))

        reschema.settings.MARKED_LOAD = False
        r = ServiceDef()
        r.load(BOOKSTORE_JSON)
        self.assertEqual(r.name, 'bookstore')
        info = r.resources['info']
        self.assertFalse(hasattr(info.input, 'start_mark'))

    def test_unknown_schema(self):
        import tempfile
        fd, name = tempfile.mkstemp(suffix='.txt', text=True)

        try:
            f = os.fdopen(fd, 'w')
            f.write('testingdata')
            f.close()

            r = ServiceDef()
            with self.assertRaises(ValueError):
                r.load(name)
        finally:
            os.unlink(name)

    def test_parse_schema(self):
        r = ServiceDef()
        with open(BOOKSTORE_YAML, 'r') as f:
            r.parse_text(f.read(), format='yaml')
        self.assertEqual(r.name, 'bookstore')

    def test_parse_schema_json(self):
        r = ServiceDef()
        with open(BOOKSTORE_JSON, 'r') as f:
            r.parse_text(f.read())
        self.assertEqual(r.name, 'bookstore')

    def test_load_bad_schema(self):
        with open(BOOKSTORE_YAML, 'r') as f:
            schema = f.readlines()
        for i, line in enumerate(schema):
            if "type: object" in line:
                schema.insert(i-1, '      bad_object_name: foo\n')
                break

        r = ServiceDef()
        with self.assertRaises(MarkedYAMLError):
            r.parse_text(''.join(schema), format='yaml')

    def test_resource_load(self):
        r = ServiceDef()
        r.load(BOOKSTORE_YAML)
        self.assertEquals(len(r.resources), 8)
        self.assertIn('info', r.resources)
        self.assertIn('author', r.resources)
        self.assertIn('authors', r.resources)
        self.assertTrue(r.find_resource('author'))
        self.assertEquals(r.find_type('no_resource'), None)

    def test_type_load(self):
        r = ServiceDef()
        r.load(BOOKSTORE_YAML)
        self.assertIn('address', r.types)
        self.assertTrue(r.find_type('address'))
        self.assertEquals(r.find_type('no_type'), None)

    def test_resource_objects(self):
        r = ServiceDef()
        r.load(BOOKSTORE_YAML)
        a = r.resources['author']
        self.assertFalse(a.is_ref())
        self.assertFalse(a.is_simple())
        self.assertIn('id', a.properties)
        self.assertIn('name', a.properties)
        self.assertEqual(a.fullid(True), '#/resources/author')
        self.assertIsNone(a.parent)
        resources = [x for x in r.resource_iter()]
        self.assertEqual(len(resources), 8)

    def test_find_name_basic(self):
        r = ServiceDef()
        r.load(BOOKSTORE_YAML)
        a = ServiceDef.find(r, '#/resources/author')
        self.assertEqual(a.fullid(True), '#/resources/author')

    def test_find_name_complex(self):
        r = ServiceDef()
        r.load(BOOKSTORE_YAML)
        c = ServiceDef.find(r, '#/resources/book/chapters')
        self.assertEqual(c.fullid(True),
                         '#/resources/book/properties/chapters')
        self.assertEqual(c._type, 'array')
        self.assertEqual(c.typestr, 'array of <object>')
        c = ServiceDef.find(r, '#/resources/book/chapters/1')
        self.assertEqual(c.fullid(True),
                         '#/resources/book/properties/chapters/items')
        self.assertEqual(c._type, 'object')
        c = ServiceDef.find(r, '#/resources/book/author_ids')
        self.assertEqual(c._type, 'array')
        c = ServiceDef.find(r, '#/resources/book/author_ids/1')
        self.assertEqual(c._type, 'integer')


class TestBookstore(unittest.TestCase):

    def setUp(self):
        self.r = ServiceDef()
        self.r.load(BOOKSTORE_YAML)

    def tearDown(self):
        self.r = None

    def test_string(self):
        s = self.r.resources['author'].properties['name']
        self.assertFalse(s.is_ref())
        self.assertTrue(s.is_simple())

        # successful validation will return None
        self.assertIsNone(s.validate('foo'))
        self.assertIsNone(s.validate(u'bar'))

        with self.assertRaises(ValidationError):
            s.validate(42)

    def test_integer(self):
        n = self.r.resources['author'].properties['id']
        self.assertFalse(n.is_ref())
        self.assertTrue(n.is_simple())

        # successful validation will return None
        self.assertIsNone(n.validate(443))

        with self.assertRaises(ValidationError):
            n.validate('foo')
        with self.assertRaises(ValidationError):
            n.validate('43')
        with self.assertRaises(ValidationError):
            n.validate(3.14)

    def test_reference(self):
        pub = ServiceDef.find(self.r, '#/resources/publisher')
        ref = pub.properties['billing_address']
        self.assertFalse(ref.is_simple())
        self.assertTrue(ref.is_ref())
        self.assertEqual(ref.typestr, 'address')

    def test_link_target(self):
        s = self.r.resources['author']
        link = s.relations['instances']
        self.assertEqual(link.resource, self.r.resources['authors'])

    def test_link_path(self):
        pub = ServiceDef.find(self.r, '#/resources/publisher')
        link = pub.links['self']
        path = link.path
        self.assertEqual(str(path), '$/publishers/{id}')
        (resolved_path, values) = path.resolve({'id': 12})
        return self.assertEqual(resolved_path, '$/publishers/12')

    def test_link_template_path(self):
        book = ServiceDef.find(self.r, '#/resources/book')
        chapters = book.properties['chapters']
        items = chapters.items
        book_chapter = items.relations['full']
        data = {'book': book.example}
        (uri, values) = book_chapter.resolve(data, '/book/chapters/1')
        self.assertEqual(uri, '$/books/100/chapters/2')
        with self.assertRaises(MissingParameter):
            book_chapter.resolve(None)

    def test_object(self):
        # skip validation, we are checking that elsewhere
        s = self.r.resources['author']
        self.assertFalse(s.is_ref())
        self.assertFalse(s.is_simple())
        self.assertIsInstance(repr(s), str)

    def test_object_xml(self):
        book = ServiceDef.find(self.r, '#/resources/book')
        p = {'id': 1,
             'title': '50 Shades of JSON',
             'publisher_id': 2,
             'author_ids': [50, 51],
             'chapters': [{'num': 1, 'heading': 'Chapter 1'},
                          {'num': 2, 'heading': 'Chapter 2'},
                          {'num': 3, 'heading': 'Chapter 3'},
                          ]}
        self.assertIsNone(book.validate(p))
        xml = book.toxml(p)
        self.assertEqual(sorted(xml.keys()),
                         [u'id', u'publisher_id', u'title'])
        xml_child = book.toxml(p, parent=xml)
        self.assertNotEqual(xml, xml_child)

    def test_array(self):
        s = self.r.resources['authors']
        self.assertFalse(s.is_ref())
        self.assertFalse(s.is_simple())

        # successful validation will return None
        p = [{'id': 1, 'name': 'Ted Nugent'},
             {'id': 2, 'name': 'Ralph Macchio'}]
        self.assertIsNone(s.validate(p))
        xml = s.toxml(p)
        self.assertEqual(len(xml.getchildren()), 2)

        with self.assertRaises(ValidationError):
            s.validate('foo')

    def test_indexing(self):
        book = self.r.resources['book']
        self.assertEqual(type(book), Object)
        self.assertEqual(type(book['id']), Integer)
        self.assertEqual(type(book['title']), String)
        self.assertGreater(len(book.str_simple()), 0)
        self.assertGreater(len(book.str_detailed()), 0)

        a = book['author_ids']
        self.assertEqual(type(a), Array)
        self.assertEqual(book.by_pointer('/author_ids'), a)

        # Test JSON pointer syntax
        self.assertEqual(type(book['author_ids'][0]), Integer)
        self.assertEqual(book.by_pointer('/author_ids/0'),
                         book['author_ids'][0])

        # Test relative JSON pointer syntax
        a0 = book.by_pointer('/author_ids/0')
        self.assertEqual(a0.by_pointer('/'), a0)
        self.assertEqual(a0.by_pointer('2/'), book)
        self.assertEqual(a0.by_pointer('2/id'), book['id'])

        with self.assertRaises(KeyError):
            book['foo']

        with self.assertRaises(KeyError):
            book.by_pointer('1/')

        with self.assertRaises(TypeError):
            a['a']

        with self.assertRaises(TypeError):
            a['10a']

        with self.assertRaises(KeyError):
            a0.by_pointer('3/')

        # Unlike book, auther does not specify additionalProperties
        author = self.r.resources['author']
        Schema.parse({}, name='<prop>', parent=author)
        self.assertEqual(author['randomjunk'], author.additional_properties)
        self.assertEqual(author.additional_properties,
                         author.children[-1])


class TestBookstoreLinks(unittest.TestCase):

    def setUp(self):
        self.r = ServiceDef()
        self.r.load(BOOKSTORE_YAML)

    def tearDown(self):
        self.r = None

    def test_links(self):
        book = self.r.resources['book']
        book_data = {'id': 1, 'title': 'My first book',
                     'publisher_id': 5, 'author_ids': [1, 5]}
        book.validate(book_data)

        author_id = book['author_ids'][0]
        logger.debug('author_id: %s' % author_id)

        (uri, values) = (author_id
                                 .relations['full']
                                 .resolve(book_data, '/author_ids/0'))

        self.assertEqual(uri, '$/authors/1')

        (uri, values) = (author_id
                                 .relations['full']
                                 .resolve(book_data, '/author_ids/1'))

        self.assertEqual(uri, '$/authors/5')

        author = self.r.resources['author']
        author_data = {'id': 1, 'name': 'John Q'}
        author.validate(author_data)

        (uri, values) = (author
                                 .relations['books']
                                 .resolve(author_data))

        logger.debug('author.relations.books uri: %s %s'
                     % (uri, values))


class TestSchemaBase(unittest.TestCase):

    def parse(self, string):
        d = yaml_loader.marked_load(string)
        s = ServiceDef(manager=ServiceDefManager())
        s.id = 'http://support.riverbed.com/apis/testschema/1.0'
        return Schema.parse(d, 'root', servicedef=s, id='#/resources/foo')

    def check_valid(self, s, valid=None, invalid=None, toxml=False):
        if type(s) is str:
            schema = self.parse(s)
        else:
            schema = s

        for a in valid:
            try:
                schema.validate(a)
            except ValidationError, e:
                self.fail("ValidationError: value should pass: %s, %s" %
                          (a, e))

            if toxml:
                schema.toxml(a)

        for a in invalid:
            try:
                schema.validate(a)
                self.fail("ValidationError not raised for value: %s" % a)
            except ValidationError:
                pass

    def check_bad_schema(self, s, etype):
        try:
            self.parse(s)
        except etype, e:
            logger.debug('Got validation error: %s' % str(e))
            return
        self.fail('Schema should have thrown error, %s' % etype)


class TestJsonSchema(TestSchemaBase):

    def setUp(self):
        self.servicedef = reschema.ServiceDef(manager=ServiceDefManager())
        self.servicedef.id = 'http://support.riverbed.com/apis/testschema/1.0'

    def test_exceptions(self):
        # cover exception string output
        s = "type: boolean\n"
        d = yaml_loader.marked_load(s)
        try:
            Schema.parse(d)
        except ValidationError, e:
            self.assertIsNotNone(str(e))

    def test_missing_servicedef(self):
        s = "type: boolean\n"
        d = yaml_loader.marked_load(s)
        with self.assertRaises(ValidationError):
            Schema.parse(d)

    def test_unnamed(self):
        s = "type: boolean\n"
        d = yaml_loader.marked_load(s)
        j = Schema.parse(d, servicedef=self.servicedef)
        self.assertTrue(j.name.startswith('element'))

    def test_unknown_type(self):
        s = "type: frobnosticator\n"
        d = yaml_loader.marked_load(s)
        with self.assertRaises(ParseError):
            Schema.parse(d)

    def test_empty(self):
        e = Schema.parse({}, servicedef=self.servicedef)
        self.check_valid(e,
                         valid=[None,
                                True,
                                False,
                                -1,
                                9999999.33333333,
                                'some string',
                                ['array', 42, {}],
                                {'object': 'stuff', 'nmber': 2}],
                         invalid=[])

    def test_boolean(self):
        self.check_valid("type: boolean\n",

                         # values to validate
                         valid=[True,
                                False],

                         invalid=["TRUE",
                                  "FALSE",
                                  "1",
                                  1]
                         )

    def test_ref_parse_error(self):
        with self.assertRaises(ParseError):
            self.parse("type: object\n"
                       "properties:\n"
                       "    id: { type: number }\n"
                       "    name: { type: string }\n"
                       "    billing_address: { type: striing }\n")

    def test_ref_invalid(self):
        schema = self.parse("type: object\n"
                            "properties:\n"
                            "    id: { type: number }\n"
                            "    name: { type: string }\n"
                            "    billing_address: { $ref: '#/address' }\n")

        with self.assertRaises(InvalidReference):
            schema.validate({'id': 2,
                             'name': 'Frozzle',
                             'billing_address': "doesn't exist"})

    def test_no_manager(self):
        schema = self.parse("type: object\n"
                            "properties:\n"
                            "    id: { type: number }\n"
                            "    name: { type: string }\n"
                            "    billing_address:"
                            " { $ref: '/api/other/1.0#/types/address' }\n")
        schema.servicedef.manager = None
        with self.assertRaises(NoManager):
            schema.validate({'id': 2,
                             'name': 'Frozzle',
                             'billing_address': "doesn't exist"})

    def test_null(self):
        self.check_valid("type: 'null'\n",

                         # values to validate
                         valid=[None],

                         invalid=["11",
                                  11,
                                  [1, 2],
                                  "abc",
                                  "123",
                                  "1234512345"],
                         )

    def test_data(self):
        # missing 'content_type'
        self.check_bad_schema("type: data\n",
                              ParseError)

        self.check_valid("type: data\n"
                         "content_type: text\n"
                         "description: simple data\n",

                         # values to validate
                         valid=["aa",
                                "11",
                                "abc",
                                "123",
                                "1234512345"],

                         invalid=[]
                         )
        schema = self.parse("type: data\n"
                            "content_type: text\n"
                            "description: simple data\n")
        self.assertTrue(schema.is_simple())

    def test_string(self):
        self.check_bad_schema("type: string\n"
                              "foobar: bad property",
                              ParseError)

        self.check_bad_schema("type: string\n"
                              "minLength: 2a",
                              ParseError)

        self.check_valid("type: string\n"
                         "minLength: 2\n"
                         "maxLength: 10\n"
                         "pattern: '^[a-z0-9]+$'\n",

                         # values to validate
                         valid=["aa",
                                "11",
                                "abc",
                                "123",
                                "1234512345"],

                         invalid=["A",
                                  "abcABCD",
                                  "12345123451"]
                         )

        self.check_valid("type: string\n"
                         "minLength: 2\n",

                         # values to validate
                         valid=["aa",
                                "1234512345" * 10],

                         invalid=["A"]
                         )

        self.check_valid("type: string\n"
                         "enum: [one, two, three]\n",

                         # values to validate
                         valid=["one",
                                "two",
                                "three"],

                         invalid=["four",
                                  "one1",
                                  "onetwo"]
                         )

        schema = self.parse("type: string\n"
                            "minLength: 2\n"
                            "maxLength: 10\n"
                            "enum: [one, two, three]\n"
                            "pattern: '[a-z0-9]+'\n"
                            "default: 'one'\n")
        self.assertIsInstance(schema.str_detailed(), basestring)

    def test_number(self):
        self.check_valid("type: number\n",
                         valid=[0, 1, 1.0, long(1), -1, -1.0, long(-1)],
                         invalid=['hi', True, False])

        self.check_valid("type: number\n"
                         "minimum: 2\n"
                         "maximum: 100\n",

                         valid=[2, 2.0, 2.5, 99, 100],
                         invalid=[1, 101, 0]
                         )

        self.check_valid("type: number\n"
                         "minimum: 2\n"
                         "maximum: 100\n"
                         "exclusiveMinimum: true\n"
                         "exclusiveMaximum: true\n",

                         valid=[2.01, 3, 99, 99.999],
                         invalid=[2, 100, 1]
                         )

        self.check_valid("type: number\n"
                         "enum: [1, 2, 3]\n",

                         valid=[1, 2, 3],
                         invalid=[0, 100, 4, 'one']
                         )

        schema = self.parse("type: number\n"
                            "minimum: 2\n"
                            "maximum: 100\n"
                            "enum: [2, 3, 4]\n"
                            "default: 3\n")
        self.assertIsInstance(schema.str_detailed(), basestring)

    def test_integer(self):
        self.check_valid("type: integer\n",
                         valid=[0, 1, long(1), -1, long(-1)],
                         invalid=[1.0, float(1), -1.0, float(-1), 'hi'])

        self.check_valid("type: integer\n"
                         "minimum: 2\n"
                         "maximum: 100\n",

                         valid=[2, 99, 100],
                         invalid=[1, 101, 2.5, 99.1]
                         )

        self.check_valid("type: integer\n"
                         "minimum: 2\n"
                         "maximum: 100\n"
                         "exclusiveMinimum: true\n"
                         "exclusiveMaximum: true\n",

                         valid=[3, 99],
                         invalid=[2, 100, 1]
                         )

        self.check_valid("type: integer\n"
                         "enum: [1, 2, 3]\n",

                         valid=[1, 2, 3],
                         invalid=[0, 100, 4, 1.0, 'one']
                         )

        schema = self.parse("type: integer\n"
                            "minimum: 2\n"
                            "maximum: 100\n"
                            "enum: [2, 3, 4]\n"
                            "default: 3\n")
        self.assertIsInstance(schema.str_detailed(), basestring)

    def test_timestamp(self):
        self.check_valid("type: timestamp\n",

                         valid=[1234567890,
                                1234567890.123,
                                long(1)],
                         invalid=['foo',
                                  {'timestamp': 1234567890},
                                  True, False]
                         )

        self.check_valid("type: timestamp-hp\n",

                         valid=[1234567890,
                                1234567890.123000,
                                long(1)],
                         invalid=['foo',
                                  {'timestamp': 1234567890},
                                  True, False]
                         )

    def test_object_simple(self):
        # required field not a list
        self.check_bad_schema("type: object\n"
                              "required: foo\n"
                              "properties:\n"
                              "  foo: { type: number }\n"
                              "  bar: { type: number }\n",
                              ParseError)

        self.check_valid("type: object\n"
                         "properties:\n"
                         "   foo: { type: number }\n"
                         "   bar: { type: string }\n"
                         "additionalProperties: false\n",

                         # values to validate
                         valid=[{"foo": 1,
                                 "bar": "one"}],

                         invalid=[{"foo": 1,
                                   "baz": "one"},
                                  'not a dict'],
                         toxml=True
                         )

    def test_object_add_props_simple(self):
        self.check_valid("type: object\n"
                         "properties:\n"
                         "   foo: { type: number }\n"
                         "   bar: { type: string }\n"
                         "additionalProperties: true",

                         # values to validate
                         valid=[{"foo": 1,
                                 "bar": "one",
                                 "baz": 2}],

                         invalid=[{"foo": "one",
                                   "baz": "one"}]
                         )

    def test_object_add_props_complex(self):
        self.check_valid("type: object\n"
                         "properties:\n"
                         "   foo: { type: number }\n"
                         "   bar: { type: string }\n"
                         "additionalProperties: { type: number }\n",

                         # values to validate
                         valid=[{"foo": 1,
                                 "bar": "one",
                                 "baz": 2}],

                         invalid=[{"foo": 1,
                                   "bar": "one",
                                   "baz": "two"}],
                         toxml=True
                         )

    def test_link(self):
        # missing self link with no path
        self.check_bad_schema("type: object\n"
                              "properties:\n"
                              "   foo: { type: number }\n"
                              "links:\n"
                              "   alink:\n"
                              "      method: GET\n",
                              ParseError)

        # missing method on non-self link
        self.check_bad_schema("type: object\n"
                              "properties:\n"
                              "   foo: { type: number }\n"
                              "links:\n"
                              "   self: { path: '$/authors' }\n"
                              "   nomethod:\n"
                              "      response: { type: integer }",
                              ParseError)

        # extra link properties
        self.check_bad_schema("type: object\n"
                              "properties:\n"
                              "   foo: { type: number }\n"
                              "links:\n"
                              '   self: { path: "$/authors" }\n'
                              "   alink:\n"
                              "      invalid: GET\n",
                              ParseError)

        # invalid link properties
        self.check_bad_schema("type: object\n"
                              "properties:\n"
                              "   foo: { type: number }\n"
                              "links:\n"
                              '   self: { path: "$/authors" }\n'
                              "   alink:\n"
                              "      invalid\n",
                              ParseError)

        schema = self.parse("type: object\n"
                            "properties:\n"
                            "   foo: { type: number }\n"
                            "links:\n"
                            '   self: { path: "$/authors" }\n')
        link = schema.links['self']
        self.assertIsInstance(repr(link), str)

    def test_empty_tags(self):
        """
        Verify tags member is added even if not specified in schema
        """

        # Base Schema-derived types
        schemas = ["type: integer",
                   "type: number",
                   "type: boolean",
                   "type: string",
                   "type: 'null'",
                   "type: timestamp",
                   "type: timestamp-hp",

                   ("type: array\n"
                    "items: { type: number }"),

                   ("type: object\n"
                    "properties:\n"
                    "   foo: { type: string }")]

        for schema in schemas:
            parsed = self.parse(schema)
            self.assertEquals(parsed.tags, {})

        # Links and relations
        schema = self.parse("type: integer\n"
                            "links:\n"
                            "   self: { path: $/foo }\n"
                            "relations:\n"
                            "   foo:\n"
                            "      resource: /foo")

        self.assertEquals(schema.links['self'].tags, {})
        self.assertEquals(schema.relations['foo'].tags, {})

        # Typeless schema fragment
        schema = self.parse("oneOf:\n"
                            "- type: integer\n"
                            "- type: 'null'")
        self.assertEquals(schema.tags, {})

    def test_tags(self):
        """
        Verify tags behavior
        """

        # tag-not-a-dict: Schema-derived types, links, and relations
        self.check_bad_schema("type: integer\n"
                              "tags: hi\n",
                              ParseError)

        self.check_bad_schema("type: integer\n"
                              "tags: [ hi ]\n",
                              ParseError)

        self.check_bad_schema("type: integer\n"
                              "links:\n"
                              "  self:\n"
                              "    path: /foo\n"
                              "    tags: [ hi ]",
                              ParseError)

        self.check_bad_schema("type: integer\n"
                              "relations:\n"
                              "  foo:\n"
                              "    resource: /foo\n"
                              "    tags: [ hi ]",
                              ParseError)

        # Base Schema-derived types
        schemas = [("type: integer\n"
                    "tags: { hi: ~, quit: bye }"),

                   ("type: number\n"
                    "tags: { hi: ~, quit: bye }"),

                   ("type: boolean\n"
                    "tags: { hi: ~, quit: bye }"),

                   ("type: string\n"
                    "tags: { hi: ~, quit: bye }"),

                   ("type: 'null'\n"
                    "tags: { hi: ~, quit: bye }"),

                   ("type: timestamp\n"
                    "tags: { hi: ~, quit: bye }"),

                   ("type: timestamp-hp\n"
                    "tags: { hi: ~, quit: bye }"),

                   ("type: array\n"
                    "tags: { hi: ~, quit: bye }\n"
                    "items:\n"
                    "    type: number"),

                   ("type: object\n"
                    "tags: { hi: ~, quit: bye }\n"
                    "properties:\n"
                    "   foo: { type: string }")]

        for schema in schemas:
            parsed = self.parse(schema)
            self.assertEquals(parsed.tags, {'hi': None, 'quit': 'bye'})

        # Links and relations
        schema = self.parse("type: integer\n"
                            "links:\n"
                            "   self:\n"
                            "      path: $/foo\n"
                            "      tags: { linktag: ~ }\n"
                            "relations:\n"
                            "   foo:\n"
                            "      resource: /foo\n"
                            "      tags: { relationtag: ~ }")

        self.assertEquals(schema.links['self'].tags, {'linktag': None})
        self.assertEquals(schema.relations['foo'].tags, {'relationtag': None})

        # Schema fragment that uses oneOf/etc to define itself as allowing
        # multiple types
        schema = self.parse("oneOf:\n"
                            "- type: integer\n"
                            "- type: 'null'\n"
                            "tags: { This is my tag: ~ }")
        self.assertEquals(schema.tags, {'This is my tag': None})


class TestSchema(TestSchemaBase):

    def setUp(self):
        self.r = reschema.ServiceDef()
        self.r.load(SERVICE_DEF_TEST)

    def tearDown(self):
        self.r = None

    def test_check_references(self):
        self.assertEqual(self.r.check_references(), [])

    def test_object_required(self):
        r = self.r.resources['test_object_required']

        self.check_valid(r,

                         valid=[{'prop_number': 1,
                                 'prop_array': [99, 98]},
                                {'prop_number': 1,
                                 'prop_string': 'foo',
                                 'prop_array': [99, 98]},
                                ],

                         invalid=[{'prop_array': [99, 98]},
                                  {'prop_array': [99, 98],
                                   'prop_string': 'foo'}])

    def test_anyof1_validation(self):
        r = self.r.resources['test_anyof1']
        self.check_valid(r,

                         valid=[{'a1': 1, 'a2': 2},
                                {'a3': 3, 'a4': 4}],

                         invalid=[{'a1': 3, 'a4': 4}])

    @pytest.mark.xfail
    def test_anyof1_indexing(self):
        r = self.r.resources['test_anyof1']

        a1 = r.by_pointer('/anyOf/0/properties/a1')
        self.assertEqual(a1.fullid(True),
                         '#/test_anyof1/anyOf/0/properties/a1')

        a2 = r['a2']
        self.assertEqual(a2.fullid(), '/test_anyof1/a2')

        a2 = ServiceDef.find(self.r, '#/resources/test_anyof1/a2')
        self.assertEqual(a2.fullid(), '#/resources/test_anyof1/a2')

    def test_anyof2_validation(self):
        r = self.r.resources['test_anyof2']
        self.check_valid(r,

                         valid=[{'a1': 1, 'a2': 5},
                                {'a1': 1, 'a2': 9},
                                {'a1': 2, 'a2': 11},
                                {'a1': 2, 'a2': 19}],

                         invalid=[{'a1': 3, 'a2': 4},
                                  {'a1': 1, 'a3': 5},
                                  {'a1': 1, 'a2': 4},
                                  {'a1': 1, 'a2': "foo"},
                                  {'a1': 1, 'a2': 10},
                                  {'a1': 2, 'a2': 4},
                                  {'a1': 2, 'a2': 5},
                                  {'a1': 2, 'a2': 9},
                                  {'a1': 2, 'a2': 29}]
                         )

    @pytest.mark.xfail
    def test_anyof2_indexing(self):
        r = self.r.resources['test_anyof2']
        a1 = r['a1']
        self.assertEqual(a1.fullid(), '/test_anyof2/a1')

        a2 = r['a2']
        self.assertEqual(a2.fullid(), '/test_anyof2/a2')

        a2 = ServiceDef.find(self.r, '/test_anyof2/a2')
        self.assertEqual(a2.fullid(), '/test_anyof2/a2')

    def test_anyof3(self):
        r = self.r.resources['test_anyof3']
        self.check_valid(r,

                         valid=[{'a1': 1, 'a2': 5},
                                {'a1': 1, 'a2': 9},
                                {'a1': 2, 'a2': 11},
                                {'a1': 2, 'a2': 19}],

                         invalid=[{'a1': 3, 'a2': 4},
                                  {'a1': 1, 'a3': 5},
                                  {'a1': 1, 'a2': 4},
                                  {'a1': 1, 'a2': "foo"},
                                  {'a1': 1, 'a2': 10},
                                  {'a1': 2, 'a2': 4},
                                  {'a1': 2, 'a2': 5},
                                  {'a1': 2, 'a2': 9},
                                  {'a1': 2, 'a2': 29}])

        a2 = ServiceDef.find(self.r, '#/resources/test_anyof3/a2')
        self.assertEqual(a2.fullid(True),
                         '#/resources/test_anyof3/properties/a2')

    def test_allof_validation(self):
        r = self.r.resources['test_allof']
        self.check_valid(r,

                         valid=[{'a1': 1,
                                 'a2': {'a21_number': 2,
                                        'a21_string': 'foo'},
                                 'a3': 'f3'},

                                {'a1': 2,
                                 'a2': {'a22_number': 2,
                                        'a22_string': 'foo',
                                        'a22_array': [1, 2, 3]},
                                 'a3': 'd1'}],

                         invalid=[{'a1': 2,
                                   'a2': {'a21_number': 2,
                                          'a21_string': 'foo'},
                                   'a3': 'f3'},

                                  {'a1': 1,
                                   'a2': {'a21_numbe': 2,
                                          'a21_string': 'foo'},
                                   'a3': 'f3'},

                                  {'a1': 1,
                                   'a2': {'a21_number': 2,
                                          'a21_string': 'foo',
                                          'a21_bad': 4},
                                   'a3': 'f3'}]
                         )

    @pytest.mark.xfail
    def test_allof_indexing(self):
        r = self.r.resources['test_allof']
        self.assertEqual(r.by_pointer('/a2/a21_string').fullid(True),
                         '#/resources/test_allof/a2/a21_string')
        self.assertEqual(r.by_pointer('/a2/a22_number').fullid(True),
                         '#/resources/test_allof/a2/a22_number')
        self.assertEqual(r.by_pointer('/a2/a22_array/0').fullid(True),
                         '#/resources/test_allof/a2/a22_array/items')

    def test_oneof(self):
        r = self.r.resources['test_oneof']
        self.check_valid(r,

                         valid=[{'a1': 1, 'a2': 21},
                                {'a1': 5, 'a2': 10}],

                         invalid=[{'a1': 1, 'a2': 2},
                                  {'a1': 5, 'a2': 20}])

    def test_not(self):
        r = self.r.resources['test_not']
        self.check_valid(r,
                         valid=[1, 2, 3, 4, 6, 9, 10],
                         invalid=[0, 5, 7, 8, 11, 12, 13, 200])

    def test_link_req_resp_defaults(self):
        r = self.r.resources['test_methods']

        self.check_valid(r.links['delete'].request,
                         valid=[None],
                         invalid=[{}, [], 0, '', ' '])
        self.check_valid(r.links['delete'].response,
                         valid=[None],
                         invalid=[{}, [], 0, '', ' '])

    def test_array_constraints(self):
        r = self.r.resources['test_array_min']

        self.check_valid(r,
                         valid=[range(10), range(15)],
                         invalid=[range(0), range(9)])

        r = self.r.resources['test_array_max']

        self.check_valid(r,
                         valid=[range(0), range(10)],
                         invalid=[range(11)])

        r = self.r.resources['test_array_min_max']

        self.check_valid(r,
                         valid=[range(10), range(20)],
                         invalid=[range(0), range(9), range(21)])

    def test_self_params(self):
        r = self.r.resources['test_self_params']
        self.check_valid(r,
                         valid=[1, 2],
                         invalid=['one', '2'])
        l = r.links['self']

        # Resolution of path params using kvs
        (uri, kvs) = l.path.resolve(kvs={'x': 1, 'y': 2, 'z': 3})
        test_uri = '$/test_self_params?x=1&y=2&z=3'
        r1 = urlparse.parse_qs(urlparse.urlsplit(uri).query)
        r2 = urlparse.parse_qs(urlparse.urlsplit(test_uri).query)
        self.assertEqual(r1, r2)

    def test_self_vars(self):
        r = self.r.resources['test_self_vars']
        self.check_valid(r,
                         valid=[1, 2],
                         invalid=['one', '2'])
        l = r.links['self']

        # Resolution of path vars using vars
        (uri, kvs) = l.path.resolve(kvs={'x': 1, 'y': 2, 'z': 3})
        self.assertEqual(uri, '$/test_self_vars?x=1&y=2&z=3')

    def test_self_buried_vars(self):
        r = self.r.resources['test_self_buried_vars']
        l = r.links['self']

        # Classic resolution of all path variables using
        # a data representation
        (uri, kvs) = l.path.resolve(
            data={'id1': 3, 'buried': {'id2': 'foo'}})
        self.assertEqual(uri, '$/test_self_buried_vars/3/foo')

        # Override of 'id2' via kvs
        (uri, kvs) = l.path.resolve(
            data={'id1': 3, 'buried': {'id2': 'foo'}},
            kvs={'id2': 'bar'})
        self.assertEqual(uri, '$/test_self_buried_vars/3/bar')

        # Override of 'id1' via kvs
        (uri, kvs) = l.path.resolve(
            data={'id1': 3, 'buried': {'id2': 'foo'}},
            kvs={'id1': 4})
        self.assertEqual(uri, '$/test_self_buried_vars/4/foo')

        # Only use kvs
        (uri, kvs) = l.path.resolve(
            kvs={'id1': 4, 'id2': 'bar'})
        self.assertEqual(uri, '$/test_self_buried_vars/4/bar')

        # Missing parameters cases
        with self.assertRaises(MissingParameter):
            (uri, values) = l.path.resolve()

        with self.assertRaises(MissingParameter):
            (uri, values) = l.path.resolve(
                data={'id1': 3, 'buried': {'id3': 'foo'}})

        with self.assertRaises(MissingParameter):
            (uri, values) = l.path.resolve(
                data={'id3': 3, 'buried': {'id2': 'foo'}})

        with self.assertRaises(MissingParameter):
            (uri, values) = l.path.resolve(
                kvs={'id1': 5})

    def test_self_vars_rel(self):
        r = self.r.resources['test_self_vars_rel']
        rel = r.relations['rel']
        (uri, values) = rel.resolve(
            data={'var_id1': 3, 'var_id2': 'foo'})
        self.assertEqual(uri, '$/test_self_buried_vars/3/foo')

        (uri, values) = rel.resolve(
            kvs={'id1': 3, 'id2': 'foo'})
        self.assertEqual(uri, '$/test_self_buried_vars/3/foo')

        (uri, values) = rel.resolve(
            data={'var_id1': 3, 'var_id2': 'foo'},
            kvs={'id1': 4})
        self.assertEqual(uri, '$/test_self_buried_vars/4/foo')

        with self.assertRaises(MissingParameter):
            (uri, values) = rel.resolve(data=None)

        with self.assertRaises(MissingParameter):
            (uri, values) = rel.resolve(data=None, kvs={'v': 3})

        with self.assertRaises(MissingParameter):
            (uri, values) = rel.resolve(data={'v': 3})

    def test_uri_params(self):
        r = self.r.resources['test_uri_params']
        l = r.links['self']

        # Classic resolution of all path variables using
        # a data representation
        (uri, kvs) = l.path.resolve(
            data={'id': 3, 'meta': {'offset': 10, 'limit': 20}})
        self.assertEqual(uri, '$/test_uri_params/3?offset=10&limit=20')

        # Missing offset or limit should not be a problem as it is optional
        (uri, kvs) = l.path.resolve(
            data={'id': 3})
        self.assertEqual(uri, '$/test_uri_params/3')

        (uri, kvs) = l.path.resolve(
            data={'id': 3, 'meta': {'offset': 10}})
        self.assertEqual(uri, '$/test_uri_params/3?offset=10')

        (uri, kvs) = l.path.resolve(
            data={'id': 3, 'meta': {'limit': 20}})
        self.assertEqual(uri, '$/test_uri_params/3?limit=20')

        # Specify limit via kvs
        (uri, kvs) = l.path.resolve(
            data={'id': 3}, kvs={'limit': 5})
        self.assertEqual(uri, '$/test_uri_params/3?limit=5')

        # Add in page limit via kvs
        (uri, kvs) = l.path.resolve(
            data={'id': 3}, kvs={'limit': 5, 'page': 7})
        self.assertEqual(uri, '$/test_uri_params/3?limit=5&page=7')

        (uri, kvs) = l.path.resolve(
            data={'id': 3}, kvs={'page': 7})
        self.assertEqual(uri, '$/test_uri_params/3?page=7')

        # Check validation
        with self.assertRaises(ValidationError):
            (uri, kvs) = l.path.resolve(
                data={'id': 3, 'meta': {'offset': '10f', 'limit': 20}},
                validate=True)

        with self.assertRaises(ValidationError):
            (uri, kvs) = l.path.resolve(
                data={'id': 3}, kvs={'page': '7'}, validate=True)

        with self.assertRaises(ValidationError):
            (uri, kvs) = l.path.resolve(
                data={'id': 3}, kvs={'page': '7'}, validate=True)

class TestSchemaMerge(TestSchemaBase):

    def setUp(self):
        self.s = ServiceDef()
        self.s.load(SERVICE_DEF_TEST)
        manager = ServiceDefManager()
        manager.add(self.s)

    def test_merge(self):
        r = self.s.find('#/resources/test_merge')
        (self.check_valid(r,
                          valid=[{'val': 10}, {'val': 14}, {'val': 20}],
                          invalid=[{'val': 9}, {'val': 21}]))

    def test_merge_source_ref(self):
        r = self.s.find('#/resources/test_merge_source_ref')
        (self.check_valid(r,
                          valid=[{'p1': 10}, {'p1': 14}, {'p1': 20}],
                          invalid=[{'p1': 9}, {'p1': 21}]))

        # Verify that the original type's input was not modified
        # by the merge operation
        t = self.s.find('#/types/type_object')
        self.assertFalse('links' in t.input)

    def test_merge_with_ref(self):
        r = self.s.find('#/resources/test_merge_with_ref')
        (self.check_valid(r,
                          valid=[{'p1': 10}, {'p1': 14}, {'p1': 20}],
                          invalid=[{'p1': 9}, {'p1': 21}]))

    def test_merge_double_ref(self):
        r = self.s.find('#/resources/test_merge_double_ref')
        (self.check_valid(r,
                          valid=[10, 14, 20],
                          invalid=[9, 21]))

    def test_merge_merge(self):
        r = self.s.find('#/resources/test_merge_merge')
        (self.check_valid(r,
                          valid=[{'val': 15}, {'val': 20}],
                          invalid=[{'val': 9}, {'val': 10},
                                   {'val': 14}, {'val': 21}]))


class TestSchemaRef(TestSchemaBase):

    def setUp(self):
        self.s1 = ServiceDef()
        self.s1.load(SERVICE_DEF_TEST)
        self.s2 = ServiceDef()
        self.s2.load(SERVICE_DEF_TEST_REF)
        manager = ServiceDefManager()
        manager.add(self.s1)
        manager.add(self.s2)
        self.assertEqual(self.s1.check_references(), [])
        self.assertEqual(self.s2.check_references(), [])

    def test_ref_type(self):
        r1 = self.s2.find('#/resources/test_ref_type')
        (self.check_valid(r1,
                          valid=[{'val': 10}, {'val': 14}, {'val': 20}],
                          invalid=[{'val': 9}, {'val': 21}]))

        r2 = r1.relations['full'].resource
        (self.check_valid(r2,
                          valid=[{'val': 10}, {'val': 14}, {'val': 20}],
                          invalid=[{'val': 9}, {'val': 21}]))

        r3 = self.s2.find('#/resources/test_ref_type_full')

        self.assertTrue(r3.matches(r2))

    def test_ref_remote_types(self):
        r = self.s2.find('#/resources/test_ref_remote_types')

        (self.check_valid(r,
                          valid=[{'prop_boolean': True,
                                  'prop_number_limits': 12},
                                 {'prop_boolean': True,
                                  'prop_number_limits': 19}],

                          invalid=[{'prop_boolean': 1,
                                    'prop_number_limits': 12},
                                   {'prop_boolean': True,
                                    'prop_number_limits': 22},
                                   {'prop_boolean': False,
                                    'prop_number_limits': 19}]))

    def test_ref_relations(self):
        item = self.s1.find('#/resources/test_item')
        ref_resource = self.s2.find('#/resources/test_ref_remote_resource')
        ref_resource_item = ref_resource.relations['item'].resource

        self.assertEqual(item, ref_resource_item)

    def test_merge_remote_ref_ref(self):
        r = self.s2.find('#/resources/test_merge_remote_ref_ref')

        self.check_valid(r,
                         valid=[{'p1': True, 'p2': 12},
                                {'p1': True, 'p2': 19}],

                         invalid=[{'p1': 1, 'p2': 12},
                                  {'p1': True, 'p2': 'foo'}])


class TestLoadHook(TestSchemaBase):

    def setUp(self):
        class Hook(object):
            service_map = {
                'http://support.riverbed.com/apis/test/1.0':
                SERVICE_DEF_TEST,

                'http://support.riverbed.com/apis/test.ref/1.0':
                SERVICE_DEF_TEST_REF
                }

            def find_by_id(self, id_):
                if id_ in self.service_map:
                    s = ServiceDef.create_from_file(self.service_map[id_])
                    return s
                else:
                    raise KeyError("Invalid id: %s" % id_)

            def find_by_name(self, name, version, provider):
                assert(provider == 'riverbed')
                sid = ('http://support.riverbed.com/apis/%s/%s' %
                       (name, version))
                return self.find_by_id(sid)

        self.manager = ServiceDefManager()
        self.manager.add_load_hook(Hook())

    def tearDown(self):
        self.manager = None

    def test_find_by_id(self):
        sid = 'http://support.riverbed.com/apis/test.ref/1.0'
        s = self.manager.find_by_id(sid)
        self.assertEqual(s.id, sid)

    def test_find_by_name(self):
        sid = 'http://support.riverbed.com/apis/test.ref/1.0'
        s = self.manager.find_by_name('test.ref', '1.0')
        self.assertEqual(s.id, sid)
        s2 = self.manager.find_by_name('test.ref', '1.0')
        self.assertEqual(s2, s)

    def test_load(self):
        s = self.manager.find_by_id(
            'http://support.riverbed.com/apis/test.ref/1.0')
        rid = ('http://support.riverbed.com/apis/test/1.0'
               '#/types/type_boolean')
        r = s.find(rid)
        self.assertEqual(r.fullid(), rid)

        rid = ('http://support.riverbed.com/apis/test.ref/1.0'
               '#/resources/test_ref_remote_types')
        r = s.find(rid)
        self.assertEqual(r.fullid(), rid)

    def test_load_on_relation(self):
        s = self.manager.find_by_id(
            'http://support.riverbed.com/apis/test.ref/1.0')
        r = s.resources['test_ref_remote_types']

        sb = r.by_pointer('/prop_boolean')
        self.assertEqual(sb.fullid(),
                         ('http://support.riverbed.com/apis/test/1.0'
                          '#/types/type_boolean'))

        r = ServiceDef.find(s, '/apis/test/1.0#/resources/test_boolean')
        self.assertEqual(r.fullid(),
                         ('http://support.riverbed.com/apis/test/1.0'
                          '#/resources/test_boolean'))


class TestServiceDef(unittest.TestCase):

    SERVICE_DEF_TEMPLATE = """
$schema: {schema_id}
id: {id}
provider: {provider}
name: {name}
version: {version}
title: {title}
tags: {tags}
"""

    def create_service(self, **kwargs):
        schema_vars = {
            'schema_id': "'http://support.riverbed.com/apis/service_def/2.2'",
            'id': "'http://support.riverbed.com/apis/test/1.0'",
            'provider': "'riverbed'",
            'name': "'test'",
            'version': "'1.0'",
            'title': "'Test REST API'",
            'tags': {}
            }
        schema_vars.update(kwargs)
        text = self.SERVICE_DEF_TEMPLATE.format(**schema_vars)
        servicedef = ServiceDef.create_from_text(text, format='yaml')
        return servicedef

    def test_good(self):
        self.create_service()

    def test_numeric_version(self):
        with self.assertRaises(ParseError):
            self.create_service(version=1.0)

    def test_service_tags(self):
        """
        Verify top-level tag behavior
        """

        service_def = self.create_service()
        self.assertEquals(service_def.tags, {})

        service_def = self.create_service(tags='{hi: ~, quit: bye}')
        self.assertEquals(service_def.tags, {'hi': None, 'quit': 'bye'})


class TestI18N(unittest.TestCase):

    def setUp(self):
        self.r = ServiceDef()
        self.r.load(BOOKSTORE_YAML)

    def tearDown(self):
        self.r = None

    def test_unicode(self):

        book = self.r.resources['book']
        book_data = {'id': 1, 'title': u'\u6d4b\u8bd5',
                     'publisher_id': 5, 'author_ids': [1, 5]}
        book.validate(book_data)


if __name__ == '__main__':
    logging.basicConfig(filename='test.log', level=logging.DEBUG)
    unittest.main()



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import os
import logging
import unittest
import subprocess

from reschema.exceptions import UnsupportedSchema
from reschema import ServiceDef

logger = logging.getLogger(__name__)

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
SERVICE_21 = os.path.join(TEST_PATH, 'service_convert_2.1.yaml')
SERVICE_22 = os.path.join(TEST_PATH, 'service_convert_2.2.yaml')

PACKAGE_PATH = os.path.dirname(TEST_PATH)

CONVERT_21_22 = os.path.join(PACKAGE_PATH, 'bin', 'convert-2.1-2.2')


class TestConvert21(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_load_schema_21(self):
        s = ServiceDef()
        with self.assertRaises(UnsupportedSchema):
            s.load(SERVICE_21)

    def test_convert_21_22(self):
        if os.path.exists(SERVICE_22):
            os.unlink(SERVICE_22)

        try:
            subprocess.check_output([CONVERT_21_22, '-f', SERVICE_21,
                                     '-o', SERVICE_22])
            s = ServiceDef()
            s.load(SERVICE_22)

            t = s.find('#/types/type3')
            t.validate({'key1': 15})

        finally:
            if os.path.exists(SERVICE_22):
                os.unlink(SERVICE_22)


if __name__ == '__main__':
    logging.basicConfig(filename='test.log', level=logging.INFO)
    unittest.main()



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

"""
The `ServiceDef` class represents a single service definition as a
collection of resources and types.  An instance of this class is
normally created by parsing a service definition file.

Manual ServiceDef creation
--------------------------

An instance may be created directly froma file:

.. code-block:: python

   >>> bookstore_def = ServiceDef.create_from_file('bookstore.yaml')

This is useful for simple stand alone services.  Once created, the
services' resources and types may be browsed the object used for
validation:

.. code-block:: python

   # Get the 'author' resource
   >>> author_res = bookstore_def.resources['author']

   # Define an author and validate that the format is correct
   >>> author = dict(id=5, name='John Doe')
   >>> author_res.validate(author)

   # If the format is incorrect, a ValidationError is thrown
   >>> author = dict(id='5', name='John Doe')
   >>> author_res.validate(author)

   ValidationError: ("author.id should be a number, got '<type 'str'>'", <servicedef.Integer 'http://support.riverbed.com/apis/bookstore/1.0#/resources/author/properties/id'>)

Using Managers
--------------

For larger projects that span multiple services, a ServiceDefManager can
be used to load service defintions on demand.  This is particularly important
when one service defintion references types or resources from another
service definition.

.. code-block:: python

   # Create a ServiceDefManager to manage service definitions The
   # CustomServiceDefLoader implements the ServiceDefLoadHook andmust
   # be defined in order to load service definitions on this system.
   >>> svcdef_mgr = ServiceDefManager()
   >>> svcdef_mgr.add_load_hook(CustomServiceDefLoader)

   # Load a service by looking for it by name/version
   >>> bookstore_def = svcdef_mgr.find_by_name('bookstore', '1.0')

"""

# System imports
import urlparse
import json
from cStringIO import StringIO
from collections import OrderedDict
import logging
import traceback

from jsonpointer import JsonPointer

# Local imports
import reschema.jsonschema as jsonschema
from reschema.jsonschema import Schema
from reschema.parser import Parser
from reschema import yaml_loader, json_loader
from reschema.exceptions import (ParseError, UnsupportedSchema, NoManager,
                                 InvalidReference, DuplicateServiceId,
                                 InvalidServiceId, InvalidServiceName,
                                 ReschemaLoadHookException)
import reschema.settings

__all__ = ['ServiceDef']
logger = logging.getLogger(__name__)


SUPPORTED_SCHEMAS = frozenset([
    "http://support.riverbed.com/apis/service_def/2.2",
])
""" The set of schema versions understood by this version of reschema """


class ServiceDefLoadHook(object):
    """ Interface for load hooks.

    See ServiceDefManager.add_hook()

    """
    def find_by_id(self, id_):
        """ Find a ServiceDef by id.

        :param id_: the globally unique URI identifying a service
            definition.
        :return: a ServiceDef instance or None

        """
        raise NotImplementedError()

    def find_by_name(self, name, version, provider):
        """ Find a ServiceDef by <name,version,provider> triplet.

        :param name: the service name
        :param version: the service version
        :param provider: the provider of the service
        :return: a ServiceDef instance or None

        """
        raise NotImplementedError()


class ServiceDefManager(object):
    """ Manager for ServiceDef instances.

    A ServiceDefManager manages loading and finding ServiceDef
    instances by id as indicated in the 'id' property
    at the top level of the schema.

    """

    def __init__(self):
        self.by_id = {}
        self.by_name = {}

        # List of hooks to call in order to load schemas for as
        # yet unknown ids
        self._load_hooks = []

    def add_load_hook(self, load_hook):
        """ Add a callable hook to load a schema by id.

        :param hook: an object that implements the ServiceDefLoadHook
            interface

        Hooks are processed in order until the first hook
        returns a ServiceDef instance.

        """
        self._load_hooks.append(load_hook)

    def clear(self):
        """ Clear all known schemas. """
        logger.info("ServiceDefManager cleared")
        self.by_id = {}
        self.by_name = {}

    def add(self, servicedef):
        """ Add a new ServiceDef instance known at the given id. """
        logger.debug("%s add: %s" % (self, servicedef.id))
        sid = servicedef.id
        if sid in self.by_id:
            if self.by_id[sid] != servicedef:
                logger.debug("ids: %s" % (self.by_id.keys()))
                raise DuplicateServiceId(sid)
            return

        self.by_id[sid] = servicedef

        fullname = (servicedef.name, servicedef.version, servicedef.provider)
        self.by_name[fullname] = servicedef

        logger.info("ServiceDefManager: registered new schema: %s, %s" %
                    (fullname, sid))
        servicedef.manager = self

    def find_by_id(self, id_):
        """ Resolve an id_ to a servicedef instance.

        If a service definition by this id is not yet in the cache, load
        hooks are invoked in order until one of them returns a instance.

        :raises InvalidServiceId: No schema found for id and could not
            be loaded

        """
        if id_ not in self.by_id:
            # Not found -- try loading via our hooks
            servicedef = None
            for hook in self._load_hooks:
                try:
                    servicedef = hook.find_by_id(id_)
                except:
                    tb = traceback.format_exc()
                    raise ReschemaLoadHookException(tb)
                if servicedef:
                    break
            if servicedef is None:
                raise InvalidServiceId(
                    "Failed to load service definition: %s" % id_)
            self.add(servicedef)
        else:
            servicedef = self.by_id[id_]

        return servicedef

    def find_by_name(self, name, version, provider='riverbed'):
        """ Resolve <provider/name/version> triplet to a servicedef instance.

        If a service definition by this full name is not yet in the
        cache, load hooks are invoked in order until one of them
        returns a instance.

        :param name: the service name
        :param version: the service version
        :param provider: the provider of the service
        :return: a ServiceDef instance

        :raises InvalidServiceId: No schema found for id and could not
            be loaded

        """
        fullname = (name, version, provider)
        if fullname not in self.by_name:
            # Not found -- try loading via our hooks
            servicedef = None
            for hook in self._load_hooks:
                try:
                    servicedef = hook.find_by_name(name, version, provider)
                except:
                    tb = traceback.format_exc()
                    raise ReschemaLoadHookException(tb)
                if servicedef:
                    break
            if servicedef is None:
                raise InvalidServiceName(
                    "Failed to load service definition: %s/%s/%s" %
                    fullname)

            self.add(servicedef)
        else:
            servicedef = self.by_name[fullname]

        return servicedef


class ServiceDef(object):
    """ Loads and represents the complete service definition

    A ServiceDef parses the top-level service definition fields and
    calls `reschema.schema.Schema.parse()` on sub-sections that
    have JSON Schema-based values.

    :param manager: The ServiceManager, if any, that is keeping track
        of this particular service definition.
    """

    def __init__(self, manager=None):
        self.manager = manager

    @classmethod
    def create_from_file(cls, filename, **kwargs):
        servicedef = ServiceDef(**kwargs)
        servicedef.load(filename)
        return servicedef

    @classmethod
    def create_from_text(cls, text, format='json', **kwargs):
        servicedef = ServiceDef(**kwargs)
        servicedef.parse_text(text, format=format)
        return servicedef

    def load(self, filename):
        """Loads and parses a JSON or YAML schema.

        Support both JSON(.json) and YAML(.yml/.yaml) file formats
        as detected by filename extensions.

        :param filename: The path to the JSON or YAML file.
        :raises ValueError: if the file has an unsupported extension.
        """

        if filename.endswith('.json'):
            format = 'json'

        elif filename.endswith(('.yml', '.yaml')):
            format = 'yaml'

        else:
            raise ValueError(
                "Unrecognized file extension, use '*.json' or '*.yaml': %s"
                % filename)

        with open(filename, 'r') as f:
            return self.load_from_stream(f, format=format)

    def load_from_stream(self, f, format='yaml'):
        """Loads and parses a JSON or YAML schema.

        Support both JSON(.json) and YAML(.yml/.yaml) file formats
        as detected by filename extensions.

        :param f: An open file object.
        :raises ValueError: if the file has an unsupported extension.
        """

        if format == 'json':
            if reschema.settings.MARKED_LOAD:
                obj = json_loader.marked_load(f)
            else:
                obj = json.load(f, object_pairs_hook=OrderedDict)

        elif format == 'yaml':
            if reschema.settings.MARKED_LOAD:
                obj = yaml_loader.marked_load(f)
            else:
                obj = yaml_loader.ordered_load(f)

        else:
            raise ValueError(
                "Unrecognized format, use 'json' or 'yaml': %s" % format)

        self.parse(obj)

    def parse_text(self, text, format='yaml'):
        """Loads and parses a schema from a string.

        :param text: The string containing the schema.
        :param format: Either 'json' (the default), 'yaml' or 'yml'.
                       This much match the format of the data in the string.
        """

        stream = StringIO(text)
        if format == 'json':
            obj = json_loader.marked_load(stream)
        elif format == 'yaml' or format == 'yml':
            obj = yaml_loader.marked_load(stream)

        return self.parse(obj)

    def parse(self, obj):
        """Parses a Python data object representing a schema.

        :param obj: The Python object containig the schema data.
        :type obj: dict
        """
        # Common properties

        with Parser(obj, '<servicdef>', self) as parser:
            parser.parse('$schema', required=True, save_as='schema')
            if self.schema not in SUPPORTED_SCHEMAS:
                raise UnsupportedSchema("Unsupported schema format: %s" %
                                        self.schema)

            parser.parse('id', required=True)
            parsed_id = urlparse.urlparse(self.id)
            if not parsed_id.netloc:
                raise ParseError("Service definition 'id' property must be a "
                                 "fully qualified URI: %s" % id)

            # Preform some preprocessing:
            #  - Expand all relative $ref targets to full absoslute references
            parser.preprocess_input(self.id)

            parser.parse('provider', required=True)
            parser.parse('name', required=True)
            parser.parse('version', required=True, types=[str, unicode])
            parser.parse('title', '')
            parser.parse('status', '')

            # 'description' is a doc property, supporting either:
            #    'description' : <string>
            #    'description' : { 'file': <filename>, 'format': <format> }
            #    'description' : { 'text': <string>, 'format': <format> }
            # where 'format' is optional and defaults to 'md'

            parser.parse('description', '')

            parser.parse('documentationLink', '')
            parser.parse('defaultAuthorization')

            self.types = OrderedDict()
            for type_ in parser.parse('types', [], save=False):
                self.types[type_] = Schema.parse(obj['types'][type_],
                                                 name=type_,
                                                 id='#/types/%s' % type_,
                                                 servicedef=self)

            self.resources = OrderedDict()
            for resource in parser.parse('resources', [], save=False):
                input_ = obj['resources'][resource]
                sch = Schema.parse(input_,
                                   name=resource,
                                   id='#/resources/%s' % resource,
                                   servicedef=self)
                self.resources[resource] = sch

            parser.parse('tasks')
            parser.parse('request_headers')
            parser.parse('response_headers')
            parser.parse('errors')
            parser.parse('tags', {}, types=dict)

    def check_references(self):
        """ Iterate through all schemas and check references.

        Check all resources and types associated with this service
        defintion and verify that all references can be properly
        resolved.  This returns an array of jsonschema.Ref instances
        that cannot be resolved.

        """

        def _check(r):

            if type(r) is jsonschema.Ref:
                try:
                    # Simply access r.refschema which will cause
                    # the refence to be checked
                    r.refschema
                except InvalidReference:
                    errors.append(r)
            else:
                for c in r.children:
                    _check(c)

        errors = []

        for r in self.resource_iter():
            _check(r)

        for r in self.type_iter():
            _check(r)

        return errors

    def resource_iter(self):
        """ Generator for iterating over all resources """
        for r in self.resources:
            yield self.resources[r]

    def type_iter(self):
        """ Generator for iterating over all types """
        for r in self.types:
            yield self.types[r]

    def find_resource(self, name):
        """ Look up and return a resource by name """
        return self.resources.get(name, None)

    def find_type(self, name):
        """ Look up and return a type by name """
        return self.types.get(name, None)

    def find(self, reference):
        """ Resolve a reference using this servicedef as a relative base

        Returns a jsonschema.Schema instance

        :param reference: string reference to resolve

        The `reference` may be one of three supported forms:

           * `<server><path>#<fragment>` - fully qualified reference

           * `<path>#<fragment>` - reference is resolved against the
             same <server> as `servicedef`.  <path> starts with '/'

           * `#<fragment>` - reference is resolved against the same
             <server> and <path> as `servicedef`

        :raises InvalidReference: `reference` does not appear to
            be to the correct syntax

        """

        parsed_reference = urlparse.urlparse(reference)
        if parsed_reference.netloc or parsed_reference.path:
            # More than just a fragment, expand the id and find the full
            # servicedef by id
            full_reference = Parser.expand_ref(self.id, reference)
            reference_id = urlparse.urldefrag(full_reference)[0]

            if reference_id == self.id:
                servicedef = self
            else:
                if self.manager is None:
                    raise NoManager(reference)
                servicedef = self.manager.find_by_id(reference_id)
        else:
            servicedef = self

        # Now that we have a servicedef, look to the fragment for:
        #   '/resource/<resource_name>/...'
        #   '/types/<type_name>/...
        p = JsonPointer(parsed_reference.fragment)
        if p.parts[0] == 'resources':
            schema = servicedef.find_resource(p.parts[1])

        elif p.parts[0] == 'types':
            schema = servicedef.find_type(p.parts[1])

        else:
            raise InvalidReference("Expected '/resources' or '/types'",
                                   reference)

        if len(p.parts) > 2 and schema is not None:
            return schema.by_pointer('/' + '/'.join(p.parts[2:]))
        else:
            return schema
